/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "kiba-utils.h"
#include "kiba-cached-icon.h"
#include "kiba-inline-pixbufs.h"

#define KIBA_CACHED_ICON_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
	KIBA_TYPE_CACHED_ICON, KibaCachedIconPrivate))

G_DEFINE_TYPE (KibaCachedIcon, kiba_cached_icon, G_TYPE_OBJECT)

enum {
	CACHED_ICON_SIZE_128 = 0,
	CACHED_ICON_SIZE_64,
	CACHED_ICON_SIZE_48,
	CACHED_ICON_SIZE_24,
	CACHED_ICON_SIZE_LAST
};
static const float
cached_sizes[CACHED_ICON_SIZE_LAST] = {
	128.0f, 64.0f, 48.0f, 24.0f
};

typedef struct {
	gchar *icon;
	gchar *uri;
	
	RsvgHandle *svg_handle;
	RsvgDimensionData dimension_data;
	cairo_surface_t *cached_surfaces[CACHED_ICON_SIZE_LAST];
} KibaCachedIconPrivate;

/// private methods

static void
draw_png_surface (cairo_t *cr,
		cairo_surface_t *surface,
		gint width,
		gint height,
		void *data)
{
	KibaCachedIcon *cached_icon = data;
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	GdkPixbuf *pixbuf = NULL;
	
	if (kiba_load_icon (priv->uri, width, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}
	
	// fallback
	pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
	g_object_unref (pixbuf);
}

static void
draw_svg_surface (cairo_t *cr,
		cairo_surface_t *surface,
		gint width,
		gint height,
		void *data)
{
	KibaCachedIcon *cached_icon = data;
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	
	float size = (float) MAX(priv->dimension_data.width,
		priv->dimension_data.height);
	cairo_scale(cr,
		((gfloat) width)/size,
		((gfloat) height)/size);
	rsvg_handle_render_cairo (priv->svg_handle, cr);
}

static void
load_surfaces(
	KibaCachedIcon *cached_icon,
	cairo_t *cr)
{
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	gint i;
	
	kiba_load_svg (
		priv->uri,
		&(priv->svg_handle),
		&(priv->dimension_data));
	if (RSVG_IS_HANDLE(priv->svg_handle)) {
		for (i = 0; i < CACHED_ICON_SIZE_LAST; i++) {
			cairo_update_surface_at_size (cr,
				&(priv->cached_surfaces[i]),
				(int) cached_sizes[i],
				(int) cached_sizes[i],
				draw_svg_surface,
				cached_icon);
		}
		g_object_unref(priv->svg_handle);
		priv->svg_handle = NULL;
	} else {
		priv->dimension_data.width = 128;
		priv->dimension_data.height = 128;
		for (i = 0; i < CACHED_ICON_SIZE_LAST; i++) {
			cairo_update_surface_at_size (cr,
				&(priv->cached_surfaces[i]),
				(int) cached_sizes[i],
				(int) cached_sizes[i],
				draw_png_surface,
				cached_icon);
		}
	}
}

static void
load_icon(
	KibaCachedIcon *cached_icon,
	const gchar *icon,
	cairo_t *cr)
{
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	gchar *uri;
	
	if (priv->icon) {
		g_free(priv->icon);
		priv->icon = NULL;
	}
	
	uri = kiba_get_uri_for_icon(icon,
		GTK_ICON_LOOKUP_FORCE_SVG, 0);
	if (!uri) {
		if (priv->uri) {
			g_free (priv->uri);
			priv->uri = NULL;
		}
		
		// initiall fallback surfaces
		load_surfaces(cached_icon, cr);
		
		return;
	}
	if (!g_str_has_suffix(uri, ".svg")) {
		g_free (uri);
		uri = kiba_get_uri_for_icon(icon,
			GTK_ICON_LOOKUP_FORCE_SVG, 48);
	}
	    
	priv->icon = g_strdup(icon);
	if (priv->uri) {
		if (strcmp(uri, priv->uri) == 0) {
			g_free (uri);
			return;
		}
		g_free (priv->uri);
	}
	priv->uri = uri;
	
	// initiall surfaces
	load_surfaces(cached_icon, cr);
}

/// public methods

void
kiba_cached_icon_get_surface(
	KibaCachedIcon *cached_icon,
	gfloat current_size,
	cairo_surface_t **surface_ret,
	gint *surface_width_ret,
	gint *surface_height_ret)
{
	g_return_if_fail(KIBA_IS_CACHED_ICON(cached_icon));
	
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	gint i, choosen = 0;
	gfloat choosen_diff = cached_sizes[0];
	
	for (i = 0; i < CACHED_ICON_SIZE_LAST; i++) {
		gfloat diff = cached_sizes[i]-current_size;
		if (diff < 0.0f) {
			continue;
		}
		
		if (diff < choosen_diff) {
			choosen = i;
			choosen_diff = diff;
		}	
	}
	
	*surface_ret = priv->cached_surfaces[choosen];
	if (priv->dimension_data.width > priv->dimension_data.height) {
		*surface_width_ret = cached_sizes[choosen];
		*surface_height_ret = ((float) priv->dimension_data.height /
			(float) priv->dimension_data.width)*cached_sizes[choosen];
	} else {
		*surface_height_ret = cached_sizes[choosen];
		*surface_width_ret = ((float) priv->dimension_data.width /
			(float) priv->dimension_data.height)*cached_sizes[choosen];
	}
}

void
kiba_cached_icon_paint(
	KibaCachedIcon *cached_icon,
	cairo_t *cr,
	gfloat size,
	gfloat alpha)
{
	g_return_if_fail(KIBA_IS_CACHED_ICON(cached_icon));
	
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	cairo_surface_t *surface = NULL;
	gint surface_width = 0, surface_height = 0;
	
	kiba_cached_icon_get_surface(
		cached_icon,
		size,
		&surface,
		&surface_width,
		&surface_height);
	
	cairo_scale(cr,
		(float) surface_width/size,
		(float) surface_height/size);
	
	cairo_set_source_surface(cr,
		surface, 0, 0);
	cairo_paint_with_alpha(cr, alpha);
}


/// GObject implementation

static void
kiba_cached_icon_dispose (GObject *g_object)
{
	KibaCachedIcon *cached_icon = KIBA_CACHED_ICON(g_object);
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	gint i;
	
	if (RSVG_IS_HANDLE(priv->svg_handle)) {
		g_object_unref(priv->svg_handle);
	}
	if (priv->icon) {
		g_free(priv->icon);
	}
	if (priv->uri) {
		g_free(priv->uri);
	}
	for (i = 0; i < CACHED_ICON_SIZE_LAST; i++) {
		priv->cached_surfaces[i] = NULL;
	}
}

/// Class Functions

static void
kiba_cached_icon_class_init (KibaCachedIconClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	gobject_class->dispose = kiba_cached_icon_dispose;

	g_type_class_add_private (class, sizeof (KibaCachedIconPrivate));
}

static void
kiba_cached_icon_init (KibaCachedIcon *cached_icon)
{
	KibaCachedIconPrivate *priv =
		KIBA_CACHED_ICON_GET_PRIVATE(cached_icon);
	gint i;
	
	priv->icon = NULL;
	priv->uri = NULL;
	priv->svg_handle = NULL;
	for (i = 0; i < CACHED_ICON_SIZE_LAST; i++) {
		priv->cached_surfaces[i] = NULL;
	}
}

KibaCachedIcon*
kiba_cached_icon_new(const gchar *icon,
		cairo_t *cr)
{
	KibaCachedIcon *cached_icon;

	cached_icon = g_object_new (KIBA_TYPE_CACHED_ICON, NULL);
	load_icon(cached_icon, icon, cr);

	return cached_icon;
}
