/*
 *  Copyright (C) 2006-2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-drawable-object.h"
#include "kiba-info-win.h"
#include "kiba-image.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-x11.h"
#include "dnd-window.h"

#include "prelight.h"
#include "shadow.h"
#include "reflection.h"
#include "activity.h"
#include "continuous-progress.h"
#include "progress.h"

#include <librsvg/rsvg.h>
#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#define KIBA_DOCK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_DOCK, KibaDockPrivate))

#define MIN_DOCK_HEIGHT    30
#define MIN_MAX_SIZE 100
#define MAX_DRAG_DISTANCE_FOR_EXEC      15.0f
#define MAX_OBJ_DISTANCE_FOR_SWITCHING  10

#define DND_MIN_DISTANCE 10.0f
#define DND_MIN_DX 0.4f
#define DND_MAX_DX 2.0f
#define DND_SWITCHING_RATE 20
// stepsize for dock scrolling
#define SCROLL_STEP   2

// preference keys
#define PREFERENCES_OBJECT_IDS_KEY "dock_objects"
#define PREFERENCES_STRUT_KEY "dock_set_strut"
#define PREFERENCES_SHAPE_KEY "dock_shape"
#define PREFERENCES_POSITION_KEY "edge_position"
#define PREFERENCES_MAX_SIZE_KEY "dock_max_size"
#define PREFERENCES_BORDER_COLOR_KEY "dock_border_rgba"
#define PREFERENCES_BORDER_WIDTH_KEY "dock_border_width"
#define PREFERENCES_CAP_KEY "dock_cap_size"
#define PREFERENCES_DRAW_KEY "dock_draw"
#define PREFERENCES_CURVES_KEY "dock_curves"
#define PREFERENCES_BACKGROUND_KEY "dock_background"
#define PREFERENCES_CURVES_COLOR_KEY "dock_curves_color"
#define PREFERENCES_GRADIENT_KEY "dock_gradient"
#define PREFERENCES_TOPLEFT_KEY "dock_corner_topleft"
#define PREFERENCES_TOPRIGHT_KEY "dock_corner_topright"
#define PREFERENCES_BOTTOMLEFT_KEY "dock_corner_bottomleft"
#define PREFERENCES_BOTTOMRIGHT_KEY "dock_corner_bottomright"
#define PREFERENCES_HEIGHT_KEY "dock_margin"
#define PREFERENCES_EDGE_RADIUS_KEY "dock_edge_radius"
#define PREFERENCES_HIDE_KEY "auto_hide_enable"
#define PREFERENCES_OBJECT_SPACE_KEY "object_space"
#define PREFERENCES_OBJECT_MARGIN_KEY "object_margin"
#define PREFERENCES_OBJECT_SIZE_KEY "dock_object_size"
#define PREFERENCES_PRELIGHT_KEY "prelight_enable"
#define PREFERENCES_REFLECTION_KEY "reflection_enable"
#define PREFERENCES_SHADOW_KEY "shadow_enable"
#define PREFERENCES_ACTIVITY_KEY "activity_enable"
#define PREFERENCES_PROGRESS_KEY "progress_enable"
#define PREFERENCES_CPROGRESS_KEY "cprogress_enable"

G_DEFINE_TYPE (KibaDock, kiba_dock, GTK_TYPE_BOX)

typedef struct _KibaDockObject KibaDockObject;

typedef struct {
	// object id
	gchar *object_id;
	KibaDockObject *dock_object;
}KibaDockObjectData;

struct _KibaDockObject {
	KibaDock *dock;
	KibaObject *object;
	KibaDockObjectData *object_data;
	gfloat last_size;
	// TRUE if the user moved the object before with dnd
	gboolean custom_position;
	// scrolling
	gfloat scroll_scale;
};

typedef struct {
	GdkWindow *event_window; /**<< pointer to the event catcher window **/

	gint scroll_counter; /**<< counter for scroll acceleration **/

	GdkRectangle mouse_anchor;
	gint move_object_id;
	GtkAllocation drag_allocation;
	GtkWidget *dragging_child; /**<< currently dragged child or NULL **/
	GtkWidget *dragging_window; /**<< pointer to the temporary dnd window **/
	gint press_count; /**<< number of button presses on the dragged object **/
	guint drag_button;

	GList *object_data; /**<< list of all added KibaDockObjectData instances **/

	GList *drawableObjects; /**<< list of all added KibaDrawableObject instances **/
	
	GList *dock_objects;

	gfloat fade_alpha; /**<< dock alpha **/
	gfloat scroll_offset; /**<< object offset for dock scrolling **/

	gboolean reposition;
	gboolean queue_update_surface;
} KibaDockPrivate;


enum {
	SIG_POSITION_CHANGED = 0,
	SIG_ORIENTATION_CHANGED,
	SIG_AUTOHIDE_TOGGLED,
	SIG_OBJECT_SIZE_CHANGED,
	SIG_HEIGHT_CHANGED,
	SIG_SIZE_CHANGED,
	SIG_DRAG_STARTED,
	SIG_KIBA_DOCK_LAST
};
static guint kiba_dock_signals[SIG_KIBA_DOCK_LAST] = { 0 };

static PrefMember dock_options[] = {
	{ PREFERENCES_OBJECT_IDS_KEY, "[]", KIBA_LIST },
	{ PREFERENCES_STRUT_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_SHAPE_KEY, "podest", KIBA_STRING },
	{ PREFERENCES_POSITION_KEY, "bottom_edge", KIBA_STRING },
	{ PREFERENCES_MAX_SIZE_KEY, "0", KIBA_INT },
	{ PREFERENCES_BORDER_COLOR_KEY, "#145BC4FF", KIBA_STRING },
	{ PREFERENCES_BORDER_WIDTH_KEY, "1.0", KIBA_DOUBLE },
	{ PREFERENCES_CAP_KEY, "5", KIBA_INT },
	{ PREFERENCES_DRAW_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_CURVES_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_BACKGROUND_KEY, "gradient", KIBA_STRING },
	{ PREFERENCES_CURVES_COLOR_KEY, "#00000033", KIBA_STRING },
	{ PREFERENCES_GRADIENT_KEY, "[#082D7688]", KIBA_LIST },
	{ PREFERENCES_TOPLEFT_KEY, "rounded", KIBA_STRING },
	{ PREFERENCES_TOPRIGHT_KEY, "rounded", KIBA_STRING },
	{ PREFERENCES_BOTTOMLEFT_KEY, "rounded", KIBA_STRING },
	{ PREFERENCES_BOTTOMRIGHT_KEY, "rounded", KIBA_STRING },
	{ PREFERENCES_HEIGHT_KEY, "30", KIBA_INT },
	{ PREFERENCES_EDGE_RADIUS_KEY, "26.0", KIBA_DOUBLE },
	{ PREFERENCES_HIDE_KEY, "FALSE", KIBA_BOOLEAN },
	{ PREFERENCES_OBJECT_SPACE_KEY, "15", KIBA_INT },
	{ PREFERENCES_OBJECT_MARGIN_KEY, "10", KIBA_INT },
	{ PREFERENCES_OBJECT_SIZE_KEY, "38", KIBA_INT },
	{ PREFERENCES_PRELIGHT_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_REFLECTION_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_SHADOW_KEY, "FALSE", KIBA_BOOLEAN },
	{ PREFERENCES_ACTIVITY_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_PROGRESS_KEY, "TRUE", KIBA_BOOLEAN },
	{ PREFERENCES_CPROGRESS_KEY, "TRUE", KIBA_BOOLEAN }
};

// property ids
enum {
	_PROP_CONFIG_FILE = 1
};

/// PRIVATE METHOIDS

static KibaDockObject*
__find_dock_object(KibaDock *dock, KibaObject *object)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaDockObject *dobject;
	GList *l;
	
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		dobject = l->data;
		if (dobject->object == object)
			return dobject;
	}
	return NULL;
}

static KibaDockObject*
__get_object_at(KibaDock *dock, gint x, gint y)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaDockObject *dobject;
	gfloat dx, dy;
	GList *l;
	
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		dobject = l->data;
		
		if (dobject->object == KIBA_OBJECT(priv->dragging_child))
			continue;
		
		dx = (dobject->object->alloc.x + dobject->object->alloc.width/2.0f) - x;
		dy = (dobject->object->alloc.y + dobject->object->alloc.height/2.0f) - y;
		
		if (sqrt(dx*dx + dy*dy) <= DND_MIN_DISTANCE) {
			return dobject;
		}
	}
	
	return NULL;
}

static KibaAllocation
__get_dock_allocation (GtkWidget *widget)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	KibaDock *dock = KIBA_DOCK(widget);
	KibaAllocation event_allocation;
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];

	event_allocation = dock->shrinked_allocation;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.height = dock->d.height;
	else
		event_allocation.width = dock->d.height;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.y += (d->valignment/2)*
			(widget->allocation.height - dock->d.height);
	else
		event_allocation.x += (d->halignment/2)*
			(widget->allocation.width - dock->d.height);

	return event_allocation;
}

static KibaDockObjectData*
__get_object_data (KibaDock *dock,
		const gchar *object_id)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GList *l;

	for (l = priv->object_data; l != NULL; l = l->next) {
		KibaDockObjectData *data = l->data;

		if (strcmp (data->object_id, object_id) == 0) {
			return data;
		}
	}

	return NULL;
}

#ifdef USE_LIBSN
static void
__sn_launch_started (KibaSn  *sn,
		GObject *object,
		gpointer data)
{
	kiba_object_set_state(KIBA_OBJECT(object), KIBA_OBJECT_WORKING);
}

static void
__sn_launch_finished (KibaSn  *sn,
		gboolean success,
		GObject *object,
		gpointer data)
{
	kiba_object_unset_state(KIBA_OBJECT(object), KIBA_OBJECT_WORKING);
}
#endif

static void
__place_objects (KibaDock *dock)
{
	GtkWidget *widget = GTK_WIDGET(dock);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	KibaAllocation child_allocation;
	const KibaLayoutData *d;
	gfloat startValue, stopValue;
	gint border_width;
	GList *l;
	
	d = &kiba_layout_data[KIBA_DOCK(widget)->d.position];
	border_width = gtk_container_get_border_width(GTK_CONTAINER(widget));

	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		child_allocation.x = (gfloat) (dock->d.cap_size + border_width) +
			dock->shrinked_allocation.x + dock->d.object_space/2.0f;
		child_allocation.y = dock->shrinked_allocation.y;
		startValue = child_allocation.x;
		stopValue =  dock->shrinked_allocation.x + dock->shrinked_allocation.width -
			(gfloat) (dock->d.cap_size + border_width) -
			dock->d.object_space/2.0f + 1.0f; // +1 for odd object spaces
	} else {
		child_allocation.x = dock->shrinked_allocation.x;
		child_allocation.y = (gfloat) (dock->d.cap_size + border_width) +
			dock->shrinked_allocation.y + dock->d.object_space/2.0f;
		startValue = child_allocation.y;
		stopValue = dock->shrinked_allocation.y + dock->shrinked_allocation.height -
			(gfloat) (dock->d.cap_size + border_width) -
			dock->d.object_space/2.0f + 1.0f; // +1 for odd object spaces
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		child_allocation.x -= priv->scroll_offset;
	} else {
		child_allocation.y -= priv->scroll_offset;
	}
	
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		KibaDockObject *dobject = l->data;
		KibaObject *object = dobject->object;
		gfloat widget_offset_x;
		gfloat widget_offset_y;
		// widget scale factors
		gfloat scale_x;
		gfloat scale_y;
		// offset to dock start/end
		gfloat coff, eoff;
		// object size
		gfloat s;
		// object coordinates, switched for vertical
		gfloat *x, *y;
		// new scale factor for scrolling
		gfloat newScale;
		// factor for y coordinate
		gfloat factor;
		// half of object size
		// objects can overlap this amount
		// without being scrolled
		gfloat size2;
		
		// set widget size
		scale_x = object->widget_scale_x;
		scale_y = object->widget_scale_y;
		scale_x *= object->cfg_scale_x;
		scale_y *= object->cfg_scale_y;
		object->alloc.width  = MAX(0.0f, ((gfloat) object->size.width)*scale_x);
		object->alloc.height = MAX(0.0f, ((gfloat) object->size.height)*scale_y);
		
		// destroy factor shrinks object space
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			child_allocation.x +=
				(0.5f*object->destroy_factor)*dock->d.object_space;
		} else {
			child_allocation.y +=
				(0.5f*object->destroy_factor)*dock->d.object_space;
		}
		
		// set object position
		widget_offset_x = object->widget_offset_x;
		widget_offset_y = object->widget_offset_y;
		object->alloc.x = child_allocation.x + widget_offset_x;
		object->alloc.y = child_allocation.y + widget_offset_y;

		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			// translate x for next object
			child_allocation.x += object->alloc.width +
				(0.5f*object->destroy_factor)*dock->d.object_space;
			// translate y for current object
			object->alloc.y += (d->valignment/2.0f)*
				(((gfloat) widget->allocation.height) - object->alloc.height) -
				(d->valignment-1.0f)*((gfloat) dock->d.object_margin);
		} else {
			// translate y for next object
			child_allocation.y += object->alloc.height +
				(0.5f*object->destroy_factor)*dock->d.object_space;
			// translate x for current object
			object->alloc.x += (d->halignment/2.0f)*
				(((gfloat) widget->allocation.width) - object->alloc.width) -
				(d->halignment-1.0f)*((gfloat) dock->d.object_margin);
		}
		
		// calculate scroll offsets
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			factor = (gfloat) (d->valignment-1);
			size2 = object->alloc.width/2.0f;
			s = object->alloc.width;
			coff = startValue - size2 - object->alloc.x;
			eoff = object->alloc.x + size2 +
				((gfloat) dock->d.object_space)*object->destroy_factor*0.5f - stopValue;
			x = &object->alloc.x;
			y = &object->alloc.y;
		} else {
			factor = (gfloat) (d->halignment-1);
			size2 = object->alloc.height/2.0f;
			s = object->alloc.height;
			coff = startValue - size2 - object->alloc.y;
			eoff = object->alloc.y + size2 +
				((gfloat) dock->d.object_space)*object->destroy_factor*0.5f - stopValue;
			x = &object->alloc.y;
			y = &object->alloc.x;
		}
		
		// set scrolled position and scale factor
		if (coff > 0) {
			*y -= factor*2.0f*sqrt(coff);
			*x = startValue - size2 + sqrt(coff);
			newScale = 1.0f - MIN(coff, 100.0f)/100.0f;
			if (object->position != KIBA_OBJECT_LEFT_OF_DOCK) {
				object->position = KIBA_OBJECT_LEFT_OF_DOCK;
				// disable events
				kiba_object_make_sensitive(object, FALSE);
				if (GDK_IS_WINDOW(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)))) {
					gdk_window_set_keep_below(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)), TRUE);
					gdk_window_lower(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)));
				}
			}
		} else if (eoff > 0) {
			*y -= factor*2.0f*sqrt(eoff);
			*x = (gfloat)stopValue + size2 - 5.0f - s - sqrt(eoff);
			newScale = 1.0f - MIN(eoff/100.0f, 1.0f);
			if (object->position != KIBA_OBJECT_RIGHT_OF_DOCK) {
				object->position = KIBA_OBJECT_RIGHT_OF_DOCK;
				// disable events
				kiba_object_make_sensitive(object, FALSE);
				if (GDK_IS_WINDOW(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)))) {
					gdk_window_set_keep_below(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)), TRUE);
					gdk_window_lower(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)));
				}
			}
		} else {
			newScale = 1.0f;
			if (object->position != KIBA_OBJECT_ON_DOCK) {
				object->position = KIBA_OBJECT_ON_DOCK;
				// enable events
				kiba_object_make_sensitive(object, TRUE);
				if (GDK_IS_WINDOW(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)))) {
					gdk_window_set_keep_below(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)), FALSE);
					gdk_window_raise(kiba_drawable_object_event_window(
						KIBA_DRAWABLE_OBJECT(object)));
				}
			}
		}
		
		if (newScale > dobject->scroll_scale+0.001f || newScale < dobject->scroll_scale-0.001f) {
			// FIXME: +/- 0.001f fixes a bug where hidden objects
			// at the end had a oscilating scale value, this caused unneeded redrawing
			dobject->scroll_scale = newScale;
			kiba_object_update_scale_factor(object);
			kiba_object_do_update_scale_factor(object);
		}
		
		// force objects to be on dock
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			object->alloc.x = MAX(dock->shrinked_allocation.x, MIN(object->alloc.x,
				dock->shrinked_allocation.x + dock->shrinked_allocation.width -
				object->alloc.width
			));
		} else {
			object->alloc.y = MAX(dock->shrinked_allocation.y, MIN(object->alloc.y,
				dock->shrinked_allocation.y + dock->shrinked_allocation.height -
				object->alloc.height
			));
		}
		
		// allocate scrolled size on the widget
		GtkAllocation allocation;
		allocation.width = object->alloc.width;
		allocation.height = object->alloc.height;
		allocation.x = object->alloc.x;
		allocation.y = object->alloc.y;
		gtk_widget_size_allocate (GTK_WIDGET(object), &allocation);
		
		kiba_object_allocation_set(object);
	}
}

/**
 * scroll scaling/alpha.
 **/
static void
__scroll_scale_func(KibaObject* object,
	gfloat* x, gfloat* y, gpointer data)
{
	KibaDockObject *dobject = data;
	if (kiba_dock_get_dragging_child(dobject->dock) != GTK_WIDGET(object)) {
		*x = dobject->scroll_scale;
		*y = dobject->scroll_scale;
	} else {
		*x = 1.0f;
		*y = 1.0f;
	}
}
static void
__scroll_scale_max_func(KibaObject* object,
	gfloat* x, gfloat* y, gpointer data)
{
	*x = 1.0f;
	*y = 1.0f;
}
static void
__scroll_alpha_func(KibaDrawableObject* object,
	gfloat* alpha, gpointer data)
{
	KibaDockObject *dobject = data;
	if (kiba_dock_get_dragging_child(dobject->dock) != GTK_WIDGET(object)) {
		*alpha = dobject->scroll_scale;
	} else {
		*alpha = 1.0f;
	}
}

static void
__dnd_max_offset_func (KibaObject *object,
	float *left, float *right,
	float *top, float *bottom,
	gpointer data)
{
	KibaDockObject *dobject = data;
	*left = 0.0f;
	*right = 0.0f;
	*top = 0.0f;
	*bottom = 0.0f;
}

static void
__queue_draw (GtkWidget *widget)
{
	KibaDock *dock = KIBA_DOCK(widget);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GdkRectangle area;
	const KibaLayoutData *d;
	GList *l;

	if (!GTK_WIDGET_REALIZED(widget))
		return;

	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		area.x = widget->allocation.x;
		area.y = (d->valignment/2)*(widget->allocation.height -
			dock->d.height) + widget->allocation.y;
		area.width = widget->allocation.width;
		area.height = dock->d.height + 1;
	}
	else {
		area.x = (d->halignment/2)*(widget->allocation.width -
			dock->d.height) + widget->allocation.x;
		area.y = widget->allocation.y;
		area.width = dock->d.height + 1;
		area.height = widget->allocation.height;
	}

	// putting damaged area in KibaWin's expose queue
	kiba_win_invalidate_area (widget->parent, area);

	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaDrawableObject *object = l->data;
		kiba_drawable_object_queue_draw(object);
	}
}

static void
__draw_surface (cairo_t *cr,
		cairo_surface_t *surface,
		gint width,
		gint height,
		void *data)
{
	KibaDock *dock = data;
	const KibaLayoutData *d =
		&kiba_layout_data[dock->d.position];

	cairo_save (cr);
		// calling custom shape draw method
		dock->shape.draw_func (dock, cr, dock->shape.draw_data);
	cairo_restore (cr);

	// drawing curves
	if (GET_FLAG (dock->flags, KIBA_DOCK_DRAW_CURVES)) {
		cairo_set_source_rgba (cr,
			dock->style.curve_color.r,
			dock->style.curve_color.g,
			dock->style.curve_color.b,
			dock->style.curve_color.a);
		cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);

		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			cairo_move_to (cr, 0, height/2);

			if (d->anchor.y == 0) {
				cairo_line_to (cr, 0, 0);
				cairo_line_to (cr, width, 0);
				cairo_line_to (cr, width, height/2);
				cairo_curve_to (cr,
					width*(4.0f/5.0f), height/16,
					width*(3.0f/5.0f), height/2,
					0, height/2);
			}
			else {
				cairo_line_to (cr, 0, height);
				cairo_line_to (cr, width, height);
				cairo_line_to (cr, width, height/2);
				cairo_curve_to (cr,
					width*(4.0f/5.0f), height/4,
					width*(3.0f/5.0f), 1.3f*height,
					0, height/2);
			}
		}
		else {
			cairo_move_to (cr, width/2, 0);

			if (d->anchor.x == 0) {
				cairo_line_to (cr, 0, 0);
				cairo_line_to (cr, 0, height);
				cairo_line_to (cr, width/2, height);
				cairo_curve_to (cr,
					width/2, height*(4.0/5.0),
					width/12, height*(3.0/5.0),
					width/2, 0);
			}
			else {
				cairo_line_to (cr, width, 0);
				cairo_line_to (cr, width, height);
				cairo_line_to (cr, width/2, height);
				cairo_curve_to (cr,
					width/4, height*(4.0/5.0),
					width, height*(3.0/5.0),
					width/2, 0);
			}
		}
		cairo_fill (cr);
	}
}


static void
__sort_objects (KibaDock *dock)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GList *l;
	
	// free old object list
	for (l = priv->dock_objects; l != NULL; l = priv->dock_objects) {
		priv->dock_objects = g_list_remove(priv->dock_objects, l->data);
	}
	g_list_free(priv->dock_objects);
	priv->dock_objects = NULL;
	
	// create new one from object data
	for (l = priv->object_data; l != NULL; l = l->next) {
		KibaDockObjectData *data = l->data;
		if (data->dock_object != NULL) {
			priv->dock_objects = g_list_append(priv->dock_objects, data->dock_object);
		}
	}
}

/**
 * handles object position changing with dnd.
 **/
static void
__check_switch_objects (KibaDock *dock, gint x, gint y)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	const KibaLayoutData *d =
		&kiba_layout_data[dock->d.position];
	KibaDockObject *drag_source;
	KibaDockObject *drag_dest;
	KibaDockObjectData *source_data, *dest_data;
	GList *needle, *l;
	
	drag_source = __find_dock_object(dock, KIBA_OBJECT(priv->dragging_child));
	drag_dest = __get_object_at(dock, x, y);
	
	if (!drag_dest) return;
	
	if (!GTK_WIDGET_SENSITIVE(drag_dest->object)) return;
	
	source_data = drag_source->object_data;
	dest_data = drag_dest->object_data;
	
	for (l = priv->object_data;
	     l->data != source_data && l->data != dest_data;
	     l = l->next);
	
	if (l->data == source_data)  {
		while ((l = l->next)->data != dest_data);
		needle = l->next;
	} else {
		needle = l;
	}
	
	priv->object_data = g_list_remove(priv->object_data, source_data);
	if (needle) {
		priv->object_data = g_list_insert_before(priv->object_data,
			needle,
			source_data);
	} else {
		priv->object_data = g_list_append(priv->object_data, source_data);
	}
	
	__sort_objects(dock);
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
}


static void
__get_max_offset(GList *features,
		const KibaLayoutData *d,
		gfloat *max_fxn, gfloat *max_fxp,
		gfloat *max_fh)
{
	GList *l;
	
	for (l = features; l != NULL; l = l->next) {
		KibaVisualFeatureObject *fobject = l->data;
		gfloat offl = 0.0f, offr = 0.0f, offt = 0.0f, offb = 0.0f;
		
		fobject->feature->get_max_offset(fobject->feature,
			fobject, &offl, &offr, &offt, &offb);
		
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			*max_fxn = MAX(*max_fxn, offt);
			*max_fxp = MAX(*max_fxp, offb);
			*max_fh = MAX(*max_fh, MAX(offl, offr));
		} else {
			*max_fxn = MAX(*max_fxn, offl);
			*max_fxp = MAX(*max_fxp, offr);
			*max_fh = MAX(*max_fh, MAX(offt, offb)); // FIXME: t only relevant for top, b for bottom
		}
	}
}

/**
 * updates dock allocation with respect to their destroy factor.
 **/
static void
__update_shrinked_allocation (KibaDock *dock)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	GtkWidget *widget = GTK_WIDGET(dock); 
	KibaWin *win = KIBA_DOCK_WIN(dock);
	GtkWidget *winWidget = GTK_WIDGET(win);
	KibaAllocation allocation, buf;
	GList *l;
	// scale space difference
	gfloat sdiff = 0.0f;
	gfloat diffx = 0.0f, diffy = 0.0f;
	gfloat *dock_area_start;
	gfloat *dock_area_size;
	gint winPoint;
	gint diff;
	gboolean allow_shrinking;
	gboolean size_changed;
	
	buf = dock->shrinked_allocation;
	
	// get difference to unshrinked dock
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		KibaDockObject *dobject = l->data;
		KibaObject *object = dobject->object;
		GtkRequisition child_requisition;
		gfloat wscale_x, wscale_y;
		
		if (!KIBA_IS_DRAWABLE_OBJECT(object))
			continue;
		
		gtk_widget_size_request (GTK_WIDGET(object), &child_requisition);
		wscale_x = object->widget_scale_x;
		wscale_y = object->widget_scale_y;
		
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			sdiff += (1.0f - object->destroy_factor)*(((gfloat) object->size.height)*
				(object->destroy_factor != 0.0f ? object->cfg_scale_y*wscale_y/object->destroy_factor : 1.0f));
		} else {
			sdiff += (1.0f - object->destroy_factor)*(((gfloat) object->size.width)*
				(object->destroy_factor != 0.0f ? object->cfg_scale_x*wscale_x/object->destroy_factor : 1.0f));
		}
		sdiff += ((gfloat) dock->d.object_space)*(1.0f - object->destroy_factor);
		
		// difference in size since last requisition
		gfloat diff, size;
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		    size = child_requisition.height;
		} else {
		    size = child_requisition.width;
		}
		diff = size - dobject->last_size;
		dobject->last_size = size;
		
		gfloat scale_x = 1.0f, scale_y = 1.0f;
		GList *li;
		for (li = object->scale_data; li != NULL; li = li->next) {
			KibaObjectScaleData *data = li->data;
			if (data->scale_widget && kiba_object_destroy_factor_func != data->scale_func) {
				gfloat func_scale_x = 1.0f;
				gfloat func_scale_y = 1.0f;
				
				data->scale_func (object, &func_scale_x, &func_scale_y, data->data);
				scale_x *= func_scale_x;
				scale_y *= func_scale_y;
			}
		}
		scale_x *= object->cfg_scale_x;
		scale_y *= object->cfg_scale_y;
		
		diffx += ((gfloat) object->size.width)*scale_x*(1.0f - object->destroy_factor);
		diffy += ((gfloat) object->size.height)*scale_y*(1.0f - object->destroy_factor);
		
		if (diff != 0.0f) {
		    // change scroll offset of the dock depending on
		    // object size changes
		    switch (object->position) {
		    case KIBA_OBJECT_ON_DOCK:
			// let the objects next to visible object grow/shrink equally
			priv->scroll_offset += diff/2.0f;
			break;
		    case KIBA_OBJECT_LEFT_OF_DOCK:
			// size change translates to offset
			priv->scroll_offset += (float) diff;
			break;
		    case KIBA_OBJECT_RIGHT_OF_DOCK:
			// scroll offset does not need to change
			break;
		    default:
			g_assert_not_reached();
		    }
		}
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
	    diff = widget->requisition.width - widget->allocation.width - diffx;
	} else {
	    diff = widget->requisition.height - widget->allocation.height - diffy;
	}
	priv->scroll_offset = MAX(0.0f, MIN(priv->scroll_offset, (gfloat) diff));
	
	dock->shrinked_allocation.x = widget->allocation.x;
	dock->shrinked_allocation.y = widget->allocation.y;
	dock->shrinked_allocation.width = widget->allocation.width;
	dock->shrinked_allocation.height = widget->allocation.height;
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		dock->shrinked_allocation.y += (gint) dock->diff_l;
		dock->shrinked_allocation.height -= (gint) (dock->diff_r + dock->diff_l);
		
		if (widget->allocation.height < widget->requisition.height) {
			// dock cant be fully size, cant shrink or grow now
			__place_objects(dock);
			allow_shrinking = FALSE;
		} else {
			dock->shrinked_allocation.height -= sdiff;
			dock_area_start = &dock->shrinked_allocation.y;
			dock_area_size = &dock->shrinked_allocation.height;
			winPoint = winWidget->allocation.y;
			allow_shrinking = TRUE;
		}
	} else {
		dock->shrinked_allocation.x += (gint) dock->diff_l;
		dock->shrinked_allocation.width -= (gint) (dock->diff_r + dock->diff_l);
		
		if (widget->allocation.width < widget->requisition.width) {
			// dock cant be fully size, cant shrink or grow now
			__place_objects(dock);
			allow_shrinking = FALSE;
		} else {
			dock->shrinked_allocation.width -= sdiff;
			dock_area_start = &dock->shrinked_allocation.x;
			dock_area_size = &dock->shrinked_allocation.width;
			winPoint = winWidget->allocation.x;
			allow_shrinking = TRUE;
		}
	}
	
	// shrink/grow with respect to heading
	if (allow_shrinking) {
		if (win->dockHeading == KIBA_DOCK_HEADING_LEFT) {
			*dock_area_start = win->dockSpace[0] + dock->diff_l - winPoint;
			if (winPoint + *dock_area_start + *dock_area_size > win->dockSpace[1] - dock->diff_r)
			    *dock_area_size = *dock_area_start - win->dockSpace[1] - dock->diff_r + winPoint;
		} else if (win->dockHeading == KIBA_DOCK_HEADING_RIGHT) {
			*dock_area_start = win->dockSpace[1] - dock->diff_r - *dock_area_size - winPoint;
			if (winPoint + *dock_area_start < win->dockSpace[0] + dock->diff_l)
			    *dock_area_start = win->dockSpace[0] + dock->diff_l - winPoint;
			if (winPoint + *dock_area_start + *dock_area_size > win->dockSpace[1] - dock->diff_r)
			    *dock_area_size = *dock_area_start - win->dockSpace[1] - dock->diff_r + winPoint;
		} else {
			*dock_area_start += sdiff/2;
		}
	}
	
	size_changed = FALSE;
	
	if (buf.width != dock->shrinked_allocation.width) {
		size_changed = TRUE;
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			kiba_dock_queue_update_surface (dock);
		}
	}
	if (buf.height != dock->shrinked_allocation.height) {
		size_changed = TRUE;
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			kiba_dock_queue_update_surface (dock);
		}
	}
	
	if (size_changed ||
	    buf.x != dock->shrinked_allocation.x ||
	    buf.y != dock->shrinked_allocation.y) {
		g_signal_emit (dock, kiba_dock_signals[SIG_SIZE_CHANGED], 0);
	}
	
	__place_objects(dock);
}

inline static void
__paint_features(GList *features,
		cairo_t *cr,
		gfloat scale_x, gfloat scale_y,
		KibaObject *object)
{
	GList *l;
	for (l = features; l != NULL; l = l->next) {
		KibaVisualFeatureObject *feature_object = l->data;
		cairo_save(cr);
			cairo_translate (cr, feature_object->allocation.x, feature_object->allocation.y);
			kiba_visual_feature_draw(
				feature_object->feature, feature_object, cr);
		cairo_restore (cr);
	}
}

inline static void
resize_features(GList *features)
{
	GList *l;
	for (l = features; l != NULL; l = l->next) {
		KibaVisualFeatureObject *feature_object = l->data;
		kiba_visual_feature_do_resize(
			feature_object->feature, feature_object);
	}
}	

static void
__update_surface (KibaDock *dock) {
	GtkWidget *widget = GTK_WIDGET(dock);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaWin *win = KIBA_WIN(widget->parent);
	const KibaLayoutData *d =
		&kiba_layout_data[dock->d.position];
	cairo_t *cr;
	
	if (!GET_FLAG (dock->flags, KIBA_DOCK_DRAWABLE)) {
		return;
	}

	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		dock->surface_data.height = dock->d.height;
		dock->surface_data.width = dock->shrinked_allocation.width;
	} else {
		dock->surface_data.width = dock->d.height;
		dock->surface_data.height = dock->shrinked_allocation.height;
	}

	cr = kiba_win_cairo_create (win);
	if (!cr) {
		return;
	}
	cairo_update_surface (cr,
		&dock->surface_data,
		__draw_surface,
		dock);
	cairo_destroy (cr);
	__queue_draw (widget);
}

static void
__expose_object(KibaDock *dock,
		KibaDrawableObject *drawable_object,
		const KibaLayoutData *d,
		cairo_t *cr)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaObject *object = KIBA_OBJECT(drawable_object);
	GtkWidget *widget = GTK_WIDGET(object);
	gboolean dragging;
	
	if (drawable_object->alpha <= 0.0f) {
		return;
	}
	
	if (priv->dragging_child == widget && GTK_IS_WINDOW(priv->dragging_window)) {
		// let the dnd window handle object rendering
		gdk_window_invalidate_rect (priv->dragging_window->window, NULL, TRUE);
		dragging = TRUE;
	} else {
		dragging = FALSE;
	}

	cairo_save(cr);
	cairo_translate (cr,
		// object position
		object->alloc.x -
		// object offset
		object->surface_offset_x +
		// content position within surface_data
		drawable_object->rendered_x,
		// object position
		object->alloc.y -
		// object offset
		object->surface_offset_y +
		// content position within surface_data
		drawable_object->rendered_y);
	
	cairo_save(cr);
		__paint_features(object->features_gimmick_back, cr,
				object->surface_scale_x, object->surface_scale_y, object);
	cairo_restore(cr);

	if (!dragging) {
		cairo_save(cr);
			// translate to ...
			cairo_translate (cr,
				// centered surface scaling
				((1.0-object->surface_scale_x)*object->alloc.width)/2.0f -
				// rendered object surface
				drawable_object->rendered_x,
				// centered surface scaling
				((1.0-object->surface_scale_y)*object->alloc.height)/2.0f -
				// rendered object surface
				drawable_object->rendered_y);
			cairo_set_source_surface(cr,
				drawable_object->rendered_object->surface, 0, 0);
			cairo_paint_with_alpha(cr, drawable_object->alpha);
		cairo_restore (cr);
	}
	
	cairo_save(cr);
		__paint_features(object->features_gimmick_front, cr,
				object->surface_scale_x, object->surface_scale_y, object);
	cairo_restore(cr);
	cairo_restore(cr);
}

/// timeouts

static gint
__scroll_counter_increase (gpointer data)
{
	KibaDockPrivate *priv = data;
	priv->scroll_counter++;
	return FALSE;
}

static gint
__scroll_counter_decrease (gpointer data)
{
	KibaDockPrivate *priv = data;
	priv->scroll_counter--;
	return FALSE;
}

static gint
__move_object (gpointer data)
{
	GtkWidget *dock = data;
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	GtkWidget *win = dock->parent;
	KibaObject *object;
	GtkAllocation child_allocation;
	gint x, y;
	gint dx, dy;
	gfloat drag_distance;
	const KibaLayoutData *d;

	g_return_val_if_fail (priv->dragging_child != NULL, FALSE);

	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	object = KIBA_OBJECT(priv->dragging_child);

	if (kiba_object_get_destroying(object)) {
		if (GTK_IS_WINDOW(priv->dragging_window)) {
			gtk_widget_destroy (priv->dragging_window);
			priv->dragging_window = NULL;
			priv->dragging_child = NULL;
			return FALSE;
		}

		priv->dragging_child = NULL;

		return FALSE;
	}

	child_allocation = priv->dragging_child->allocation;

	gdk_window_get_pointer (GTK_WIDGET(dock)->window, &x, &y, NULL);
	
	// dnd between objects
	__check_switch_objects (KIBA_DOCK(dock), x, y);

	if (!GTK_IS_WINDOW(priv->dragging_window)) {
		dx = priv->dragging_child->allocation.x - x + priv->mouse_anchor.x;
		dy = priv->dragging_child->allocation.y - y + priv->mouse_anchor.y;
		drag_distance = sqrt (dx * dx + dy * dy);

		if (drag_distance > MAX_DRAG_DISTANCE_FOR_EXEC) {
			priv->dragging_window = kiba_dnd_window_new(object,
				priv->mouse_anchor.x/object->alloc.width,
				priv->mouse_anchor.y/object->alloc.height);
			kiba_object_set_state(object, KIBA_OBJECT_DRAGGING);
			// TODO: enable dnd to other docks
		}
	} if (GTK_IS_WINDOW(priv->dragging_window)) {
		kiba_dnd_window_set_mouse_position(
				priv->dragging_window,
				x + win->allocation.x, y + win->allocation.y);

		if (KIBA_IS_DRAWABLE_OBJECT(priv->dragging_child)) {
			kiba_drawable_object_queue_draw(KIBA_DRAWABLE_OBJECT(priv->dragging_child));
		}
 	}

	return TRUE;
}


/// signal handler

static void
parent_realized(GtkWidget *widget,
		gpointer data)
{
	KibaWin *win = KIBA_WIN(widget);
	KibaDock *dock = data;
	
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_REFLECTIONS)) {
		kiba_reflection_feature_new(win);
	}
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_PRELIGHT)) {
		kiba_prelight_feature_new(win, dock->config_file);
	}
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_SHADOW)) {
		kiba_shadow_feature_new(win, dock->config_file);
	}
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_ACTIVITY)) {
		kiba_activity_feature_new(win);
	}
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_CONTINUOUS_PROGRESS)) {
		kiba_continuous_progress_feature_new(win);
	}
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_PROGRESS)) {
		kiba_progress_feature_new(win);
	}
}

static gboolean
__object_button_press_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	GtkWidget *dock = data;
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	gboolean before = TRUE;
	KibaObject *object = KIBA_OBJECT(widget);
	KibaDockObject *dobject, *buf;
	gint x, y;
	gint fX, fY;
	GList *l;
	
	if (object->position != KIBA_OBJECT_ON_DOCK) {
		return TRUE;
	}
	
	if (event->button == 3) {
		if (GTK_IS_WINDOW(priv->dragging_window)) {
			if (event->type == GDK_BUTTON_PRESS)
				priv->press_count++;
			return TRUE;
		}
		return FALSE;
	}

	dobject = __find_dock_object(KIBA_DOCK(dock), object);
	priv->dragging_child = widget;

	kiba_object_hide_title (object);

	gdk_window_get_pointer (dock->window, &x, &y, NULL);

	fX = (gint) object->alloc.x;
	fY = (gint) object->alloc.y;

	priv->mouse_anchor.x = x - fX;
	priv->mouse_anchor.y = y - fY;

	if (event->type == GDK_BUTTON_PRESS)
		priv->press_count++;
	if (!priv->drag_button)
		priv->drag_button = event->button;

	if (event->button != 1) {
		return TRUE;
	}
	
	if (!priv->move_object_id) {
		priv->move_object_id = g_timeout_add (20, __move_object, dock);
	}

	return TRUE;
}

static gboolean
__object_button_release_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaDock *dock = KIBA_DOCK(data);
	KibaObject *object = KIBA_OBJECT(widget);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	gint x, y, dx, dy;
	gfloat drag_distance;
	const KibaLayoutData *d =
		&kiba_layout_data[dock->d.position];

	priv->press_count--;
	if (priv->press_count > 0 && event->button != priv->drag_button) {
		return TRUE;
	}
	priv->press_count = 0;
	priv->drag_button = 0;

	if (priv->move_object_id) {
		g_source_remove (priv->move_object_id);
		priv->move_object_id = 0;
	}

	gboolean dragged;
	if (GTK_IS_WINDOW(priv->dragging_window)) {
		gtk_widget_destroy (priv->dragging_window);
		priv->dragging_window = NULL;
		dragged = TRUE;
	} else {
		dragged = FALSE;
	}
	priv->dragging_child = NULL;
	
	if (GET_FLAG (object->state, KIBA_OBJECT_DRAGGING)) {
		kiba_object_unset_state(object, KIBA_OBJECT_DRAGGING);
	}
	
	if (dragged) {
		KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
		
		if (object->position != KIBA_OBJECT_ON_DOCK) {
			kiba_object_make_sensitive(object, FALSE);
		}
		
		return TRUE;
	}
	
	if (object->position != KIBA_OBJECT_ON_DOCK) {
		return TRUE;
	}

	gdk_window_get_pointer (GTK_WIDGET(dock)->window, &x, &y, NULL);
	x -= priv->mouse_anchor.x;
	y -= priv->mouse_anchor.y;

	dx = widget->allocation.x - x;
	dy = widget->allocation.y - y;
	drag_distance = sqrt (dx * dx + dy * dy);

	if (drag_distance < MAX_DRAG_DISTANCE_FOR_EXEC) {
		kiba_object_show_title (object);
	}
	__place_objects (dock);

	return drag_distance > MAX_DRAG_DISTANCE_FOR_EXEC;
}

static gboolean
__object_enter_notify_cb (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaDock *dock = KIBA_DOCK(data);
	KibaObject *object = KIBA_OBJECT(widget);
	
	if (GET_FLAG(object->state, KIBA_OBJECT_DRAGGING)) {
		// propagade event further
		return FALSE;
	}
	
	gtk_container_set_focus_child (GTK_CONTAINER(dock), widget);
	
	kiba_object_set_state (object, KIBA_OBJECT_PRELIGHTED);
	kiba_object_show_title (object);
	
	// propagade event further
	return FALSE;
}

static gboolean
__object_leave_notify_cb (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaObject *object = KIBA_OBJECT(widget);
	
	kiba_object_hide_title (object);
	if (!GET_FLAG(object->state, KIBA_OBJECT_DRAGGING)) {
		kiba_object_unset_state (object, KIBA_OBJECT_PRELIGHTED);
	}

	// propagade event further
	return FALSE;
}


static gboolean
__object_scroll_event (GtkWidget *widget,
		GdkEventScroll *event,
		gpointer data)
{
	return TRUE;
}

static void
__object_id_changed_cb(GObject *g_object,
		GParamSpec *pspec,
		gpointer data)
{
	KibaDock *dock = data;
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaObject *object = KIBA_OBJECT(g_object);
	KibaDockObjectData *object_data;
	gchar *object_id = NULL;
	GList *l;
	
	g_object_get (g_object,
		"object-id", &object_id, NULL);
	
	object_data = NULL;
	for (l = priv->object_data; l != NULL; l = l->next) {
		KibaDockObjectData *data = l->data;
		if (data->dock_object && data->dock_object->object == object) {
			object_data = data;
			break;
		}
	}
	g_return_if_fail (object_data != NULL);
	
	g_free(object_data->object_id);
	object_data->object_id = g_strdup(object_id);
	g_free(object_id);
}	

/// core intern methods

void
kiba_dock_set_shape (KibaDock *dock,
		KibaDockShape shape)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	dock->shape = shape;
	kiba_dock_queue_update_surface (dock);
}

void
kiba_dock_set_max_size (KibaDock *dock,
		gint max_size)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GList *l;
	
	if (max_size < MIN_MAX_SIZE) {
		const KibaLayoutData *layout;

		layout = &kiba_layout_data[dock->d.position];
		if (layout->orientation == GTK_ORIENTATION_VERTICAL)
			max_size = KIBA_SCREEN_DEFAULT->geometry.height;
		else
			max_size = KIBA_SCREEN_DEFAULT->geometry.width;
	}
	dock->d.max_size = max_size;
	
	// some objects may be shrinked
	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaDrawableObject *object = l->data;
		kiba_drawable_object_queue_update_surface(object);
	}

	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
	kiba_dock_queue_update_surface (dock);
}

void
kiba_dock_set_layout (KibaDock *dock,
		KibaDockLayout  d)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	GtkWidget *widget = GTK_WIDGET(dock);
	gboolean height_changed;

	if (d.height != dock->d.height)
		height_changed = TRUE;

	dock->d = d;

	if (height_changed) {
		g_signal_emit (dock, kiba_dock_signals[SIG_HEIGHT_CHANGED], 0, dock->d.height);
	}

	// apply new layout to the dock
	gtk_widget_queue_resize_no_redraw (GTK_WIDGET(dock));
	if (widget->parent) {
		if (height_changed) {
			gtk_widget_queue_draw(widget->parent);
		}
		gtk_widget_queue_resize_no_redraw (widget->parent);
		kiba_dock_queue_update_surface (dock);
	}
}

void
kiba_dock_set_style (KibaDock *dock,
		KibaDockStyle style)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	dock->style = style;
	kiba_dock_queue_update_surface (dock);
}

void
kiba_dock_set_flag (KibaDock *dock,
		gint flag)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	if (GET_FLAG(dock->flags, flag))
		return;

	SET_FLAG(dock->flags, flag);

	switch (flag) {
		case KIBA_DOCK_AUTOHIDE:
			g_signal_emit (dock, kiba_dock_signals[SIG_AUTOHIDE_TOGGLED],
				0, TRUE);
			break;
		case KIBA_DOCK_DRAW_CURVES:
			kiba_dock_queue_update_surface (dock);
			break;
		case KIBA_DOCK_DRAWABLE:
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_SET_STRUT:
			kiba_dock_update_strut (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_REFLECTIONS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_reflection_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent));
			}
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_SHADOW:
			if (GTK_WIDGET(dock)->parent) {
				kiba_shadow_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent), dock->config_file);
			}
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_PRELIGHT:
			if (GTK_WIDGET(dock)->parent) {
				kiba_prelight_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent), dock->config_file);
			}
			break;
		case KIBA_DOCK_DRAW_ACTIVITY:
			if (GTK_WIDGET(dock)->parent) {
				kiba_activity_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent));
			}
			break;
		case KIBA_DOCK_DRAW_CONTINUOUS_PROGRESS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_continuous_progress_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent));
			}
			break;
		case KIBA_DOCK_DRAW_PROGRESS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_progress_feature_new(KIBA_WIN(GTK_WIDGET(dock)->parent));
			}
			break;
	}
}

void
kiba_dock_unset_flag (KibaDock *dock,
	gint flag)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	if (!GET_FLAG(dock->flags, flag))
		return;

	DEL_FLAG(dock->flags, flag);

	switch (flag) {
		case KIBA_DOCK_AUTOHIDE:
			g_signal_emit (dock, kiba_dock_signals[SIG_AUTOHIDE_TOGGLED],
				0, FALSE);
			break;
		case KIBA_DOCK_DRAW_CURVES:
			kiba_dock_queue_update_surface (dock);
			break;
		case KIBA_DOCK_DRAWABLE:
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_SET_STRUT:
			kiba_dock_update_strut (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_REFLECTIONS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					REFLECTION_FEATURE_NAME);
			}
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_SHADOW:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					SHADOW_FEATURE_NAME);
			}
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_DRAW_PRELIGHT:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					PRELIGHT_FEATURE_NAME);
			}
			break;
		case KIBA_DOCK_DRAW_ACTIVITY:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					ACTIVITY_FEATURE_NAME);
			}
			break;
		case KIBA_DOCK_DRAW_CONTINUOUS_PROGRESS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					CONTINUOUS_PROGRESS_FEATURE_NAME);
			}
			break;
		case KIBA_DOCK_DRAW_PROGRESS:
			if (GTK_WIDGET(dock)->parent) {
				kiba_win_remove_visual_feature(KIBA_WIN(GTK_WIDGET(dock)->parent),
					"progress");
			}
			break;
	}
}

void
kiba_dock_set_alpha (KibaDock *dock,
		gfloat alpha)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock))->fade_alpha = alpha;
	__queue_draw (GTK_WIDGET(dock));
}

void
kiba_dock_set_object_size (KibaDock *dock,
		gint size)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GtkWidget *widget = GTK_WIDGET(dock);
	GList *l;

	dock->d.object_size = size;

	if (!GTK_WIDGET_REALIZED(dock))
		return;

	g_signal_emit (dock,
		kiba_dock_signals[SIG_OBJECT_SIZE_CHANGED],
		0, dock->d.object_size);

	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		gtk_widget_queue_resize_no_redraw (child->widget);
		if (KIBA_IS_DRAWABLE_OBJECT(child->widget)) {
			kiba_drawable_object_queue_update_surface(
				KIBA_DRAWABLE_OBJECT(child->widget));
		}
	}

	gtk_widget_queue_resize_no_redraw (widget);
}

void
kiba_dock_set_object_list (KibaDock *dock,
		const GSList *object_list)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaDockObjectData *object_data;
	GSList *l;
	
	for (l = (GSList*) object_list; l != NULL; l = l->next) {
		const gchar *object_id = l->data;
		object_data = __get_object_data(dock, object_id);
		if (!object_data) {
			object_data = g_new0 (KibaDockObjectData, 1);
			object_data->object_id = g_strdup (object_id);
			object_data->dock_object = NULL;
			priv->object_data = g_list_append(
				priv->object_data, object_data);
		}
	}
	__sort_objects(dock);
}

/**
 * Returns newly allocated list of object ids in order.
 **/
GList*
kiba_dock_get_object_list (KibaDock *dock, gboolean active_only)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	KibaDockObjectData *object_data;
	GList *l, *ret = NULL;
	
	for (l = priv->object_data; l != NULL; l = l->next) {
		object_data = l->data;
		if (!active_only || KIBA_IS_OBJECT(object_data->dock_object)) {
			ret = g_list_append(ret, g_strdup(object_data->object_id));
		}
	}
	
	return ret;
}

cairo_pattern_t*
kiba_dock_get_pattern (KibaDock *dock)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	cairo_pattern_t *ret;
	gint border_width, x0, x1, y0, y1;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));

	x0 = border_width/2;
	y0 = border_width/2;
	x1 = dock->surface_data.width  - border_width/2;
	y1 = dock->surface_data.height - border_width/2;

	if (dock->style.file) {
		ret = cairo_get_png_pattern (dock->style.file, 1.0);
	} else {
		if (dock->style.numColors == 0) {
			KibaRGBA gradient[] = {
				(KibaRGBA) {1.0f, 1.0f, 1.0f, 1.0f}
			};
			ret = cairo_get_gradient_pattern (x0, y0, x1, y1,
				gradient, 1);
		} else {
			ret = cairo_get_gradient_pattern (x0, y0, x1, y1,
				dock->style.gradient, dock->style.numColors);
		}

	}

	return ret;
}

/// Public Methods

/**
 * kiba_dock_get_scroll_offset:
 * @dock: the KibaDock instance.
 *
 * returns object offset due to dock scrolling
 *
 * Returns: offset in pixel
 */
gfloat
kiba_dock_get_scroll_offset(KibaDock *dock)
{
	g_return_val_if_fail(KIBA_IS_DOCK(dock), 0.0f);
	
	return KIBA_DOCK_GET_PRIVATE (dock)->scroll_offset;
}

/**
 * kiba_dock_get_objects:
 * @dock: the KibaDock instance.
 *
 * returns newly allocated list of KibaObject instances
 *
 * Returns: list of objects
 */
GList* kiba_dock_get_objects(KibaDock *dock)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GList *ret = NULL, *l;
	
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		KibaDockObject *dobject = l->data;
		ret = g_list_append(ret, dobject->object);
	}
	
	return ret;
}

/**
 * kiba_dock_get_dragging_child:
 * @dock: the KibaDock instance.
 *
 * returns currently dragged object
 *
 * Returns: the dragged child or NULL
 */
GtkWidget*
kiba_dock_get_dragging_child(KibaDock *dock)
{
	g_return_if_fail(KIBA_IS_DOCK(dock));
	return KIBA_DOCK_GET_PRIVATE (dock)->dragging_child;
}

/**
 * kiba_dock_get_dragging_window:
 * @dock: the KibaDock instance.
 *
 * returns dragging window
 *
 * Returns: the dragging window or NULL
 */
GtkWidget*
kiba_dock_get_dragging_window(KibaDock *dock)
{
	g_return_if_fail (KIBA_IS_DOCK(dock));
	return KIBA_DOCK_GET_PRIVATE (dock)->dragging_window;
}

/*** Render Methods ***/

/**
 * kiba_dock_expose:
 * @dock: the KibaDock instance.
 * @cr: the drawing context.
 *
 * Recursively renders the dock and its children.
 */
void
kiba_dock_expose (KibaDock *dock, cairo_t  *cr)
{
	GList *l, *objects = NULL;
	GtkWidget *win;
	const KibaLayoutData *d;
	KibaDockPrivate *priv;
	gint border_width;
	GtkWidget *focus_child;
	gfloat alpha;

	g_return_if_fail (KIBA_IS_DOCK(dock));

	win = GTK_WIDGET(dock)->parent;
	border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	d = &kiba_layout_data[dock->d.position];

	// update dock surface if queued
	if (priv->queue_update_surface) {
		priv->queue_update_surface = FALSE;
		__update_surface(dock);
	}
	
	// rendering the focussed child first
	focus_child = GTK_CONTAINER(dock)->focus_child;
	if (KIBA_IS_DRAWABLE_OBJECT(focus_child)) {
		priv->drawableObjects = g_list_remove (priv->drawableObjects, focus_child);
		priv->drawableObjects = g_list_append (priv->drawableObjects, focus_child);
	}
	
	for (l = priv->dock_objects; l != NULL; l = l->next) {
		KibaDockObject *child = l->data;
		KibaObject *object = child->object;
		if (object->position != KIBA_OBJECT_LEFT_OF_DOCK)
			break;
		if (!KIBA_IS_DRAWABLE_OBJECT(object))
			continue;
		objects = g_list_append(objects, object);
		kiba_drawable_object_update_surface(KIBA_DRAWABLE_OBJECT(object), KIBA_WIN(win));
	}
	for (l = g_list_last(priv->dock_objects); l != NULL; l = l->prev) {
		KibaDockObject *child = l->data;
		KibaObject *object = child->object;
		if (object->position != KIBA_OBJECT_RIGHT_OF_DOCK)
			break;
		if (!KIBA_IS_DRAWABLE_OBJECT(object))
			continue;
		objects = g_list_append(objects, object);
		kiba_drawable_object_update_surface(KIBA_DRAWABLE_OBJECT(object), KIBA_WIN(win));
	}
	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaObject *object = l->data;
		kiba_drawable_object_do_update_alpha(KIBA_DRAWABLE_OBJECT(object));
		if (object->position != KIBA_OBJECT_ON_DOCK)
			continue;
		objects = g_list_append(objects, object);
		kiba_drawable_object_update_surface(KIBA_DRAWABLE_OBJECT(object), KIBA_WIN(win));
	}
	
	alpha = priv->fade_alpha*dock->style.background_alpha;
	KibaAllocation dockAllocation = __get_dock_allocation(GTK_WIDGET(dock));
	if (GET_FLAG (dock->flags, KIBA_DOCK_DRAWABLE) && alpha > 0.0f) {
		cairo_save (cr);

		cairo_save (cr);
		cairo_translate (cr, dockAllocation.x, dockAllocation.y);
		
		// paint below dock features
		for (l = objects; l != NULL; l = l->next) {
			KibaObject *object = l->data;

			cairo_save(cr);
				cairo_translate (cr,
					object->alloc.x - object->surface_offset_x,
					object->alloc.y - object->surface_offset_y);
				__paint_features(object->features_ontop_dock_back, cr,
						object->surface_scale_x, object->surface_scale_y, object);
			cairo_restore(cr);
		}

		// drawing the Dock
		cairo_save (cr);
			cairo_set_source_surface (cr, dock->surface_data.surface, 0, 0);
			cairo_paint_with_alpha (cr, alpha);
		cairo_restore (cr);
		cairo_restore (cr);
		cairo_save(cr);

		cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);

		if (dock->shape.clip_func) {
			dock->shape.clip_func (dock, cr, dock->shape.draw_data);
		}
		
		// paint above dock features
		for (l = objects; l != NULL; l = l->next) {
			KibaObject *object = l->data;
			
			cairo_save(cr);
				cairo_translate (cr,
					object->alloc.x - object->surface_offset_x,
					object->alloc.y - object->surface_offset_y);
				__paint_features(object->features_ontop_dock_front, cr,
					object->surface_scale_x, object->surface_scale_y, object);
			cairo_restore (cr);
		}

		cairo_restore (cr);
	}
	
	// drawing the objects
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	for (l = objects; l != NULL; l = objects) {
		KibaObject *object = l->data;
		__expose_object(dock,
			KIBA_DRAWABLE_OBJECT(object),
			d, cr);
		objects = g_list_remove(objects, object);
	}
	g_list_free(objects);
}

/**
 * kiba_dock_set_position:
 * @dock: the KibaDock instance.
 * @position: the position to set.
 *
 * Sets dock position.
 */
void
kiba_dock_set_position (KibaDock *dock,
		KibaPosition position)
{
	const KibaLayoutData *old_d;
	const KibaLayoutData *new_d;
	KibaDockPrivate *priv;
	GList *l;

	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (position >= 0);

	priv = KIBA_DOCK_GET_PRIVATE (dock);

	if (position == dock->d.position) {
		return;
	}

	old_d = &kiba_layout_data[dock->d.position];
	new_d = &kiba_layout_data[position];

	dock->d.position = position;
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
	kiba_dock_queue_update_surface(dock);
	kiba_dock_queue_reposition(dock);
	if (old_d->orientation != new_d->orientation)
		g_signal_emit (dock,
			kiba_dock_signals[SIG_ORIENTATION_CHANGED],
			0, new_d->orientation);
	g_signal_emit (dock,
		kiba_dock_signals[SIG_POSITION_CHANGED],
		0, position);
}

/**
 * kiba_dock_get_position:
 * @dock: the KibaDock instance.
 *
 * Returns dock position.
 *
 * Returns: the dock position.
 */
KibaPosition
kiba_dock_get_position (KibaDock *dock)
{
	g_return_val_if_fail (KIBA_IS_DOCK(dock), KIBA_POSITION_BOTTOM_EDGE);
	return dock->d.position;
}

/**
 * kiba_dock_update_strut:
 * @widget: the KibaDock instance.
 *
 * Sets/updates _NET_WM_STRUT property on the dock window.
 * This will prevent other windows to overlap with the dock in maximized state.
 */
void
kiba_dock_update_strut (GtkWidget *widget)
{
	KibaStrut strut;
	KibaDock *dock;
	const KibaLayoutData *d;

	if (!GTK_WIDGET_REALIZED(widget))
		return;

	dock = KIBA_DOCK(widget);
	d = &kiba_layout_data[dock->d.position];

	if (!GET_FLAG(dock->flags, KIBA_DOCK_SET_STRUT) ||
	     GET_FLAG(dock->flags, KIBA_DOCK_AUTOHIDE) ||
	     GET_FLAG(KIBA_WIN(widget->parent)->probs, KIBA_WIN_BELOW)) {
		strut.left = 0;
		strut.right = 0;
		strut.top = 0;
		strut.bottom = 0;
	}
	else if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		strut.left = 0;
		strut.right = 0;
		strut.top = dock->d.height * ((2-d->valignment)/2);
		strut.bottom = dock->d.height * (d->valignment/2);
	}
	else {
		strut.left = dock->d.height * ((2-d->halignment)/2);
		strut.right = dock->d.height * (d->halignment/2);
		strut.top = 0;
		strut.bottom = 0;
	}

	kiba_set_wm_strut (gdk_x11_drawable_get_xid (widget->window), &strut);
}

/**
 * kiba_dock_add_populate_func:
 * @dock: the KibaDock instance.
 * @func: the method to populate the dock.
 * @icon: the menu item icon path or name.
 * @title: the menu item title.
 * @user_data: The user data.
 *
 * Adds method to populate the dock with objects.
 */
void
kiba_dock_add_populate_func (KibaDock *dock,
		KibaDockPopulateFunc func,
		const gchar *icon,
		const gchar *title,
		gpointer user_data)
{
	KibaDockPopulateData *data;
	GList *l;

	g_return_if_fail (dock != NULL);
	g_return_if_fail (func != NULL);
	g_return_if_fail (icon != NULL);
	g_return_if_fail (title != NULL);

	for (l = dock->populate_data; l != NULL; l = l->next) {
		data = l->data;

		if (data->func == func) {
			data->user_data = data;
			return;
		}
	}

	data = g_new0 (KibaDockPopulateData, 1);
	data->dock = dock;
	data->func = func;
	data->icon = g_strdup(icon);
	data->title = g_strdup(title);
	data->user_data = user_data;

	dock->populate_data = g_list_append (dock->populate_data, data);
}

/**
 * kiba_dock_remove_populate_func:
 * @dock: the KibaDock instance.
 * @func: the method to populate the dock.
 *
 * Removes method to populate the dock with objects.
 */
void
kiba_dock_remove_populate_func (KibaDock *dock,
		KibaDockPopulateFunc func)
{
	GList *l;

	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (func != NULL);

	for (l = dock->populate_data; l != NULL; l = l->next) {
		KibaDockPopulateData *data = l->data;

		if (data->func == func) {
			dock->populate_data = g_list_remove (dock->populate_data, data);
			g_free (data->icon);
			g_free (data->title);
			g_free (data);
			return;
		}
	}
}

/**
 * kiba_dock_queue_reposition:
 * @dock: the KibaDock instance.
 *
 * Queues dock repositioning for the next loop
 */
void
kiba_dock_queue_reposition (KibaDock *dock)
{
	g_return_if_fail (KIBA_IS_DOCK(dock));
	KIBA_DOCK_GET_PRIVATE (dock)->reposition = TRUE;
}

/**
 * kiba_dock_queue_update_surface:
 * @dock: the KibaDock instance.
 *
 * Queues a surface update for the next loop
 */
void
kiba_dock_queue_update_surface (KibaDock *dock)
{
	g_return_if_fail (KIBA_IS_DOCK(dock));
	KIBA_DOCK_GET_PRIVATE (dock)->queue_update_surface = TRUE;
	__queue_draw(GTK_WIDGET(dock));
}

gchar*
kiba_dock_get_name(KibaDock *dock)
{
	g_return_val_if_fail(KIBA_IS_DOCK(dock), NULL);
	return kiba_prefs_get_name(dock->config_file);
}

/// preferences callbacks

static void
preferences_changed_cb (const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaDock *dock = data;
	
	if (strcmp(option_id, PREFERENCES_OBJECT_IDS_KEY) == 0) {
	} else if (strcmp(option_id, PREFERENCES_GRADIENT_KEY) == 0) {
		KibaDockStyle style = dock->style;
		GSList *list = value;
		gint length = g_slist_length(list);
		KibaRGBA *gradient = g_new0(KibaRGBA, length);
		kiba_prefs_string_list_to_rgba_array(list, gradient);
		if (style.gradient)
			g_free (style.gradient);
		style.gradient = gradient;
		style.numColors = length;
		kiba_dock_set_style (dock, style);
	} else if (strcmp(option_id, PREFERENCES_BACKGROUND_KEY) == 0) {
		KibaDockStyle style = dock->style;
		if (g_str_has_suffix ((gchar*) value, ".png") ||
		    g_str_has_suffix ((gchar*) value, ".svg")) {
			style.file = (gchar*) value;
		} else {
			style.file = NULL;
		}
		kiba_dock_set_style (dock, style);
	} else if (strcmp(option_id, PREFERENCES_CURVES_COLOR_KEY) == 0) {
		KibaDockStyle style = dock->style;
		style.curve_color = kiba_prefs_rgba_from_string ((gchar*) value);
		kiba_dock_set_style (dock, style);
	} else if (strcmp(option_id, PREFERENCES_BORDER_COLOR_KEY) == 0) {
		KibaDockStyle style = dock->style;
		style.border_rgba = kiba_prefs_rgba_from_string ((gchar*) value);
		kiba_dock_set_style (dock, style);
	} else if (strcmp(option_id, PREFERENCES_HEIGHT_KEY) == 0) {
		KibaDockLayout d = dock->d;
		d.height = GPOINTER_TO_INT (value);
		kiba_dock_set_layout (dock, d);
	} else if (strcmp(option_id, PREFERENCES_SHAPE_KEY) == 0) {
		if (strcmp((gchar*) value, "podest") == 0) {
			kiba_dock_podest_shape_activate (dock);
		} else if (strcmp((gchar*) value, "default") == 0) {
			kiba_dock_default_shape_activate (dock);
		}
	} else if (strcmp(option_id, PREFERENCES_POSITION_KEY) == 0) {
		kiba_dock_set_position (dock,
			kiba_prefs_get_position_from_string((gchar*) value));
	} else if (strcmp(option_id, PREFERENCES_MAX_SIZE_KEY) == 0) {
		kiba_dock_set_max_size (dock, GPOINTER_TO_INT (value));
	} else if (strcmp(option_id, PREFERENCES_BORDER_WIDTH_KEY) == 0) {
		gtk_container_set_border_width (GTK_CONTAINER(dock),
			(gint) (gfloat)g_ascii_strtod((gchar*) value, NULL));
	} else if (strcmp(option_id, PREFERENCES_CAP_KEY) == 0) {
		KibaDockLayout d = dock->d;
		d.cap_size = GPOINTER_TO_INT (value);
		kiba_dock_set_layout (dock, d);
	} else if (strcmp(option_id, PREFERENCES_STRUT_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_SET_STRUT);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_SET_STRUT);
		}
	} else if (strcmp(option_id, PREFERENCES_DRAW_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAWABLE);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAWABLE);
		}
	} else if (strcmp(option_id, PREFERENCES_CURVES_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_CURVES);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_CURVES);
		}
	} else if (strcmp(option_id, PREFERENCES_PRELIGHT_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_PRELIGHT);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_PRELIGHT);
		}
	} else if (strcmp(option_id, PREFERENCES_REFLECTION_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_REFLECTIONS);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_REFLECTIONS);
		}
	} else if (strcmp(option_id, PREFERENCES_SHADOW_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_SHADOW);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_SHADOW);
		}
	} else if (strcmp(option_id, PREFERENCES_ACTIVITY_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_ACTIVITY);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_ACTIVITY);
		}
	} else if (strcmp(option_id, PREFERENCES_PROGRESS_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_PROGRESS);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_PROGRESS);
		}
	} else if (strcmp(option_id, PREFERENCES_CURVES_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_CURVES);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_CURVES);
		}
	} else if (strcmp(option_id, PREFERENCES_CPROGRESS_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_DRAW_CONTINUOUS_PROGRESS);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_DRAW_CONTINUOUS_PROGRESS);
		}
	} else if (strcmp(option_id, PREFERENCES_HIDE_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_dock_set_flag (dock, KIBA_DOCK_AUTOHIDE);
		} else {
			kiba_dock_unset_flag (dock, KIBA_DOCK_AUTOHIDE);
		}
	} else if (strcmp(option_id, PREFERENCES_TOPLEFT_KEY) == 0) {
		dock->shape.topleft =  kiba_prefs_edge_style_from_string ((gchar*) value);
		kiba_dock_set_shape (dock, dock->shape);
	} else if (strcmp(option_id, PREFERENCES_BOTTOMLEFT_KEY) == 0) {
		dock->shape.bottomleft =  kiba_prefs_edge_style_from_string ((gchar*) value);
		kiba_dock_set_shape (dock, dock->shape);
	} else if (strcmp(option_id, PREFERENCES_TOPRIGHT_KEY) == 0) {
		dock->shape.topright =  kiba_prefs_edge_style_from_string ((gchar*) value);
		kiba_dock_set_shape (dock, dock->shape);
	} else if (strcmp(option_id, PREFERENCES_BOTTOMRIGHT_KEY) == 0) {
		dock->shape.bottomright =  kiba_prefs_edge_style_from_string ((gchar*) value);
		kiba_dock_set_shape (dock, dock->shape);
	} else if (strcmp(option_id, PREFERENCES_EDGE_RADIUS_KEY) == 0) {
		dock->shape.corner_radius = (gfloat)g_ascii_strtod((gchar*) value, NULL);
		kiba_dock_set_shape (dock, dock->shape);
	} else if (strcmp(option_id, PREFERENCES_OBJECT_SPACE_KEY) == 0) {
		KibaDockLayout d = dock->d;
		d.object_space = GPOINTER_TO_INT (value);
		kiba_dock_set_layout (dock, d);
	} else if (strcmp(option_id, PREFERENCES_OBJECT_MARGIN_KEY) == 0) {
		KibaDockLayout d = dock->d;
		d.object_margin = GPOINTER_TO_INT (value);
		kiba_dock_set_layout (dock, d);
	} else if (strcmp(option_id, PREFERENCES_OBJECT_SIZE_KEY) == 0) {
		kiba_dock_set_object_size (dock, GPOINTER_TO_INT(value));
	} else {
		DEBUG_ERROR("Unknow KibaDock option: %s\n", option_id);
	}
}

/// Class Implementations

/**
 * gtk_widget_parent_set implementation.
 **/
static void
kiba_dock_parent_set(GtkWidget *widget,
		GtkWidget *old_parent)
{
	KibaDock *dock = KIBA_DOCK(widget);
	    
	if (KIBA_IS_WIN(old_parent)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(old_parent),
			G_CALLBACK(parent_realized), dock);
	}
	if (widget->parent) {
		g_signal_connect_after (G_OBJECT(widget->parent), "realize",
			G_CALLBACK(parent_realized), dock);
		if (GTK_WIDGET_REALIZED(widget->parent)) {
			parent_realized(widget->parent, dock);
		}
	}
}

/**
 * gtk_container_add implementation.
 **/
static void
kiba_dock_add (GtkContainer *container,
		GtkWidget *widget)
{
	KibaDock *dock = KIBA_DOCK(container);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (container);
	KibaObject *object;
	KibaDockObject *dobject;
	KibaDockObjectData *object_data;
	gchar *object_id;
	GList *l;

	g_return_if_fail (KIBA_IS_OBJECT(widget));

	object = KIBA_OBJECT(widget);
	
	// get object data
	g_object_get (G_OBJECT(object), "object-id", &object_id, NULL);
	object_data = __get_object_data(dock, object_id);
	if (object_data != NULL)
		g_return_if_fail(object_data->dock_object == NULL);
	
	// add dock object
	dobject = g_new0(KibaDockObject, 1);
	dobject->object = KIBA_OBJECT(gtk_widget_ref(widget));
	dobject->dock = KIBA_DOCK(gtk_widget_ref(GTK_WIDGET(dock)));
	dobject->last_size = 0.0f;
	dobject->scroll_scale = 1.0f;
	dobject->custom_position = FALSE;
	priv->dock_objects = g_list_prepend(priv->dock_objects, dobject);
	
	if (KIBA_IS_DRAWABLE_OBJECT(object)) {
		priv->drawableObjects = g_list_append(priv->drawableObjects, object);
		kiba_drawable_object_add_alpha_function (KIBA_DRAWABLE_OBJECT(object),
			__scroll_alpha_func, dobject);
	}

	// connecting object signal handler
	g_signal_connect (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(__object_button_press_cb), container);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(__object_button_release_cb), container);
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(__object_enter_notify_cb), container);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(__object_leave_notify_cb), container);
	g_signal_connect_after (G_OBJECT(widget), "scroll-event",
		G_CALLBACK(__object_scroll_event), container);
	g_signal_connect (G_OBJECT(widget), "notify::object-id",
		G_CALLBACK(__object_id_changed_cb), container);
	
	// add scroll scaling method
	kiba_object_add_scale_function (object,
		__scroll_scale_func,
		__scroll_scale_max_func,
		FALSE, dobject);
	
	if (object_data == NULL) {
		object_data = g_new0 (KibaDockObjectData, 1);
		object_data->object_id = g_strdup (object_id);
		priv->object_data = g_list_append(priv->object_data, object_data);
	}
	object_data->dock_object = dobject;
	g_free(object_id);
	// set object data attr
	dobject->object_data = object_data;
	
	(* GTK_CONTAINER_CLASS (kiba_dock_parent_class)->add) (container, widget);
	
	// sorting objects
	__sort_objects(dock);
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
}

/**
 * gtk_container_remove implementation.
 **/
static void
kiba_dock_remove (GtkContainer *container,
		GtkWidget *widget)
{
	KibaDock *dock = KIBA_DOCK(container);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (container);
	KibaObject *object = KIBA_OBJECT(widget);
	KibaDockObject *dobject;
	
	// delete dock object
	dobject = __find_dock_object(dock, object);
	g_return_if_fail(dobject != NULL);
	
	// remove dobject from object data
	dobject->object_data->dock_object = NULL;
	
	// free dobject
	priv->dock_objects = g_list_remove(priv->dock_objects, dobject);
	gtk_widget_unref(GTK_WIDGET(dobject->dock));
	gtk_widget_unref(GTK_WIDGET(dobject->object));
	g_free(dobject);
	
	// remove from drawable list
	if (g_list_find(priv->drawableObjects, object))
		priv->drawableObjects = g_list_remove(priv->drawableObjects, object);
	
	(* GTK_CONTAINER_CLASS (kiba_dock_parent_class)->remove) (container, widget);
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(dock));
}

/**
 * gtk_widget_scroll_event implementation.
 **/
static gboolean
kiba_dock_scroll_event (GtkWidget *widget,
		GdkEventScroll *event)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	KibaDock *dock = KIBA_DOCK(widget);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	gint diff;

	// moving objects to the left
	if (event->direction == GDK_SCROLL_UP) {
		priv->scroll_counter++;
		priv->scroll_counter = MIN(priv->scroll_counter, 15);
		g_timeout_add (100, __scroll_counter_decrease, priv);
	}
	// moving objects to the right
	else if (event->direction == GDK_SCROLL_DOWN) {
		priv->scroll_counter--;
		priv->scroll_counter = MAX(priv->scroll_counter, -15);
		g_timeout_add (100, __scroll_counter_increase, priv);
	}
	priv->scroll_offset += priv->scroll_counter*SCROLL_STEP;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
	    diff = widget->requisition.width-widget->allocation.width; // FIXME: wrong!
	else
	    diff = widget->requisition.height-widget->allocation.height; // FIXME: wrong!
	priv->scroll_offset = MAX(0.0f, MIN(priv->scroll_offset, (gfloat) diff));

	__place_objects (dock);

	return TRUE;
}

/**
 * gtk_widget_size_request implementation
 **/
static void
kiba_dock_size_request (GtkWidget *widget,
		GtkRequisition *requisition)
{
	KibaDock *dock = KIBA_DOCK(widget);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	gint border_width = gtk_container_get_border_width(
		GTK_CONTAINER(widget));
	gint num_objects = g_list_length(priv->dock_objects);
	gfloat max_ox = 0.0f, max_oh = 0.0f;
	gfloat max_fxn = 0.0f, max_fxp = 0.0f, max_fh = 0.0f;
	gfloat max_diff = 0.0f;
	gfloat diff_l, diff_r;
	GList *l;
	
	requisition->width = 0;
	requisition->height = 0;
	
	struct {
		gint *w, *h;
		gint alignment;
	} magic[2] = {
		{ &requisition->width,  &requisition->height, d->valignment},
		{ &requisition->height, &requisition->width,  d->halignment }
	};
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object = KIBA_OBJECT(child->widget);
		
		kiba_object_do_update_scale_factor(object);
		kiba_object_do_update_offset(object);
	}
	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaDrawableObject *dobject = l->data;
		kiba_drawable_object_do_update_area(dobject);
	}
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object = KIBA_OBJECT(child->widget);
		
		resize_features(object->features_morph_front);
		resize_features(object->features_morph_back);
		resize_features(object->features_ontop_front);
		resize_features(object->features_ontop_back);
		resize_features(object->features_gimmick_front);
		resize_features(object->features_gimmick_back);
		resize_features(object->features_ontop_dock_front);
		resize_features(object->features_ontop_dock_back);
	}
	
	
	
	// make sure object animations can be displayed
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object = KIBA_OBJECT(child->widget);
		GList *li;
		
		gfloat widget_scale_x = 1.0f, widget_scale_y = 1.0f;
		gfloat surface_scale_x = 1.0f, surface_scale_y = 1.0f;
		gfloat o_offset_x = 0.0f, o_offset_y = 0.0f;
		// maximal widgets size
		gfloat ww, wh;
		// difference from widget scaling to surface scaling
		gfloat dx, dy;
		
		for (li = object->scale_data; li != NULL; li = li->next) {
			KibaObjectScaleData *data = li->data;
			gfloat custom_factor_x = 1.0f;
			gfloat custom_factor_y = 1.0f;

			data->max_func (object,
				&custom_factor_x,
				&custom_factor_y,
				data->data);
			
			if (data->scale_widget) {
				widget_scale_x *= custom_factor_x;
				widget_scale_y *= custom_factor_y;
			} else {
				surface_scale_x *= custom_factor_x;
				surface_scale_y *= custom_factor_y;
			}
		}
		widget_scale_x *= object->cfg_scale_x;
		widget_scale_y *= object->cfg_scale_y;
		
		for (li = object->offset_data; li != NULL; li = li->next) {
			KibaObjectOffsetData *data = li->data;
			gfloat custom_offset_l = 0.0f;
			gfloat custom_offset_r = 0.0f;
			gfloat custom_offset_b = 0.0f;
			gfloat custom_offset_t = 0.0f;

			data->max_func (object,
				&custom_offset_l,
				&custom_offset_r,
				&custom_offset_b,
				&custom_offset_t,
				data->data);
			
			o_offset_x += custom_offset_r+custom_offset_l;
			o_offset_y += custom_offset_b+custom_offset_t;
		}
		
		ww = widget_scale_x*object->size.width;
		wh = widget_scale_y*object->size.height;
		dx = ww*(surface_scale_x - 1.0f);
		dy = wh*(surface_scale_y - 1.0f);
		
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			max_oh = MAX(max_oh, ww + o_offset_x);
			max_ox = MAX(max_ox, dy + 2.0f*o_offset_y);
			requisition->height += (wh*object->cfg_scale_y)/widget_scale_y;
		} else {
			max_oh = MAX(max_oh, wh + o_offset_y);
			max_ox = MAX(max_ox, dx + 2.0f*o_offset_x);
			requisition->width += (ww*object->cfg_scale_x)/widget_scale_x;
		}
		
		// get max feature offset
		__get_max_offset(object->features_gimmick_front,
			d, &max_fxn, &max_fxp, &max_fh);
		__get_max_offset(object->features_gimmick_back,
			d, &max_fxn, &max_fxp, &max_fh);
		__get_max_offset(object->features_morph_front,
			d, &max_fxn, &max_fxp, &max_fh);
		__get_max_offset(object->features_morph_back,
			d, &max_fxn, &max_fxp, &max_fh);
	}
	
	diff_l = (gint) (0.5f*max_ox + max_fxn)+1;
	diff_r = (gint) (0.5f*max_ox + max_fxp)+1;
	if (diff_l != dock->diff_l || diff_r != dock->diff_r) {
		gtk_widget_queue_resize_no_redraw(widget->parent);
		kiba_dock_queue_reposition(dock);
	}
	dock->diff_l = diff_l;
	dock->diff_r = diff_r;
	
	*magic[d->orientation].w += 2*border_width +
		MAX(num_objects-1,0)*dock->d.object_space +
		2.0f*(dock->d.cap_size + dock->d.object_space) + diff_l + diff_r;
	*magic[d->orientation].h = dock->d.object_margin +
		border_width + max_oh + max_fh +
		10; // FIXME: 10 is enough for scroll animation, but not nice
	
	if (*magic[d->orientation].w < KIBA_DOCK_MIN_SIZE) {
		*magic[d->orientation].w = KIBA_DOCK_MIN_SIZE;
	}

	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (requisition->width > KIBA_SCREEN_DEFAULT->geometry.width)
			requisition->width = KIBA_SCREEN_DEFAULT->geometry.width;
	} else {
		if (requisition->height > KIBA_SCREEN_DEFAULT->geometry.height)
			requisition->height = KIBA_SCREEN_DEFAULT->geometry.height;
	}
}

/**
 * gtk_widget_size_allocate implementation.
 **/
static void
kiba_dock_size_allocate (GtkWidget *widget,
		GtkAllocation *allocation)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	KibaDock *dock = KIBA_DOCK(widget);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	GtkAllocation event_allocation;
	gboolean size_changed, position_changed;
	
	size_changed = (widget->allocation.width != allocation->width ||
	    widget->allocation.height != allocation->height);
	position_changed = (widget->allocation.x != allocation->x ||
	    widget->allocation.y != allocation->y);

	if (size_changed || position_changed) {
		__queue_draw (GTK_WIDGET(dock));
		widget->allocation = *allocation;
		__queue_draw (GTK_WIDGET(dock));
	}
	
	__place_objects (dock);

	if (!GTK_WIDGET_REALIZED (widget)) {
		return;
	}
	
	if (size_changed) {
		gdk_window_move_resize (priv->event_window,
				widget->allocation.x,
				widget->allocation.y,
				widget->allocation.width,
				widget->allocation.height);
	} else {
		gdk_window_move (priv->event_window,
				widget->allocation.x,
				widget->allocation.y);
	}
	
	if (priv->reposition) {
		priv->reposition = FALSE;
		__update_shrinked_allocation(dock);
	} else if (size_changed) {
		__update_shrinked_allocation(dock);
	}
	
	// updating the _NET_WM_STRUT property
	kiba_dock_update_strut (widget);
}

/**
 * gtk_widget_unmap implementation.
 **/
static void
kiba_dock_unmap (GtkWidget *widget) {
	gdk_window_hide (KIBA_DOCK_GET_PRIVATE (widget)->event_window);

	// calling the GtkWidget implementation
	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->unmap) (widget);
}

/**
 * gtk_widget_map_implemantation.
 **/
static void
kiba_dock_map (GtkWidget *widget) {
	// calling the GtkWidget implementation
	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->map) (widget);

	gdk_window_show (KIBA_DOCK_GET_PRIVATE (widget)->event_window);
}

/**
 * gtk_widget_unrealize implementaion.
 **/
static void
kiba_dock_unrealize (GtkWidget *widget)
{
	KibaDock *dock = KIBA_DOCK(widget);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	GList *l;

	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->unrealize) (widget);
	
#ifdef USE_LIBSN
	KibaSn *sn = kiba_sn_get();
	g_signal_handlers_disconnect_by_func (G_OBJECT(sn),
		G_CALLBACK(__sn_launch_finished), widget);
	g_signal_handlers_disconnect_by_func (G_OBJECT(sn),
		G_CALLBACK(__sn_launch_started), widget);
#endif

	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;

	if (dock->surface_data.surface) {
		cairo_surface_destroy (dock->surface_data.surface);
	}
	dock->surface_data.surface = NULL;

	if (priv->move_object_id) {
		g_source_remove (priv->move_object_id);
	}
	priv->move_object_id = 0;
	
	for (l = priv->object_data; l != NULL; l = priv->object_data) {
		KibaDockObjectData *data = l->data;
		
		priv->object_data = g_list_remove (priv->object_data, data);
		g_free (data->object_id);
		g_free (data);
	}
	g_list_free (priv->object_data);
	priv->object_data = NULL;
}

/**
 * gtk_widget_realize implementation.
 **/
static void
kiba_dock_realize (GtkWidget *widget)
{
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	GdkWindowAttr attributes;
	gint attributes_mask;

	// make sure the dock is properly size allocated
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 1, 1 };
		gtk_widget_size_request (widget, &requisition);
		gtk_widget_size_allocate (widget, &allocation);
	}

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window);

	// creating a event catcher window
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;

	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	// catching the window events on the widget
	gdk_window_set_user_data (priv->event_window, widget);
	// allowing dnd on the window
	gdk_window_register_dnd (priv->event_window);
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);

	widget->style = gtk_style_attach (widget->style, widget->window);

#ifdef USE_LIBSN
	KibaSn *sn = kiba_sn_get();
	g_signal_connect(G_OBJECT(sn), "launch-finished",
		G_CALLBACK(__sn_launch_finished), widget);
	g_signal_connect(G_OBJECT(sn), "launch-started",
		G_CALLBACK(__sn_launch_started), widget);
#endif
}

/**
 * g_object_set_property implementation.
 **/
static void
kiba_dock_set_property (GObject *object,
		guint property_id,
		const GValue *value,
		GParamSpec *pspec)
{
	KibaDock *self = KIBA_DOCK(object);

	switch (property_id) {
		case _PROP_CONFIG_FILE:
			self->config_file = g_value_get_pointer (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/**
 * g_object_get_property implementation.
 **/
static void
kiba_dock_get_property (GObject *object,
		guint property_id,
		GValue *value,
		GParamSpec *pspec)
{
	KibaDock *self = KIBA_DOCK(object);

	switch (property_id) {
		case _PROP_CONFIG_FILE:
			g_value_set_pointer (value, self->config_file);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

static void
kiba_dock_class_init (KibaDockClass *class)
{
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

	g_type_class_add_private (class, sizeof (KibaDockPrivate));

	// GObject method implementations
	gobject_class->set_property = kiba_dock_set_property;
	gobject_class->get_property = kiba_dock_get_property;

	// custom widget implementations
	widget_class->realize = kiba_dock_realize;
	widget_class->unrealize = kiba_dock_unrealize;
	widget_class->map = kiba_dock_map;
	widget_class->unmap = kiba_dock_unmap;
	widget_class->size_request = kiba_dock_size_request;
	widget_class->size_allocate = kiba_dock_size_allocate;
	widget_class->scroll_event = kiba_dock_scroll_event;
	widget_class->parent_set = kiba_dock_parent_set;

	// custom container implementations
	container_class->add = kiba_dock_add;
	container_class->remove = kiba_dock_remove;
	
	/**
	 * KibaDock:config-file:
	 *
	 * Configuration file containing dock specific settings.
	 */
	g_object_class_install_property
		(gobject_class,
		 _PROP_CONFIG_FILE,
		 g_param_spec_pointer ("config_file",
		 "Config File",
		 "Configuration file",
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	
	/**
	 * KibaDock::position-changed:
	 * @dock: the object on which the signal is emitted
	 * @position: the new dock position.
	 *
	 * The position-changed signal is emitted when a new position has been set
	 * on @dock.
	 */
	kiba_dock_signals[SIG_POSITION_CHANGED] =
		g_signal_new ("position-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, position_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	/**
	 * KibaDock::position-changed:
	 * @dock: the object on which the signal is emitted
	 * @orientation: the new dock orientation.
	 *
	 * The orientation-changed signal is emitted when a new orientation has been set
	 * on @dock.
	 */
	kiba_dock_signals[SIG_ORIENTATION_CHANGED] =
		g_signal_new ("orientation-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, orientation_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	/**
	 * KibaDock::autohide-toggled:
	 * @dock: the object on which the signal is emitted
	 * @autohide: autohide active?
	 *
	 * The autohide-toggled signal is emitted when autohiding
	 * was togfled.
	 */
	kiba_dock_signals[SIG_AUTOHIDE_TOGGLED] =
		g_signal_new ("autohide-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, autohide_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaDock::object-size-changed:
	 * @dock: the object on which the signal is emitted
	 * @object_size: the new object size
	 *
	 * The object-size-changed signal is emitted when the
	 * object size was set.
	 */
	kiba_dock_signals[SIG_OBJECT_SIZE_CHANGED] =
		g_signal_new ("object-size-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, object_size_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	/**
	 * KibaDock::height-changed:
	 * @dock: the object on which the signal is emitted
	 * @height: the new dock height
	 *
	 * The height-changed signal is emitted when the
	 * dock height was set.
	 */
	kiba_dock_signals[SIG_HEIGHT_CHANGED] =
		g_signal_new ("height-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, height_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	/**
	 * KibaDock::size-changed:
	 * @dock: the object on which the signal is emitted
	 *
	 * The size-changed signal is emitted when the
	 * dock size changed.
	 */
	kiba_dock_signals[SIG_SIZE_CHANGED] =
		g_signal_new ("size-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, height_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0,
			G_TYPE_NONE);
}

static void
kiba_dock_init (KibaDock *dock)
{
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};

	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);

	// using no widget window
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (dock), GTK_NO_WINDOW);
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (dock), GTK_DOUBLE_BUFFERED);
	
	gtk_widget_set_redraw_on_allocate (GTK_WIDGET(dock), FALSE);
	gtk_widget_set_app_paintable (GTK_WIDGET (dock), TRUE);
	
	gtk_widget_add_events (GTK_WIDGET (dock),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	gtk_container_set_border_width (GTK_CONTAINER(dock), 5);
	
	dock->d.position = KIBA_POSITION_BOTTOM_EDGE;
	dock->d.cap_size = 5;
	dock->d.object_space = 10;
	dock->d.object_margin = 10;
	dock->d.object_size = 38;
	dock->d.height = 25;
	dock->d.max_size = 0;
	dock->style.draw = TRUE;
	dock->style.file = NULL;
	dock->style.background_alpha = 1.0f;
	dock->style.border_rgba = (KibaRGBA) { 0.6f, 0.6f, 0.6f, 0.7f };
	dock->style.curve_color = (KibaRGBA) { 0.0f, 0.0f, 0.0f, 0.3f };
	dock->style.gradient = NULL;
	dock->style.numColors = 0;
	dock->d.max_size = KIBA_SCREEN_DEFAULT->geometry.width;
	dock->surface_data.surface = NULL;
	dock->shape.draw_func = NULL;
	dock->shape.clip_func = NULL;
	dock->shape.draw_data = NULL;
	dock->shape.topleft = KIBA_ROUNDED;
	dock->shape.topright = KIBA_ROUNDED;
	dock->shape.bottomright = KIBA_ROUNDED;
	dock->shape.bottomleft = KIBA_ROUNDED;
	dock->shape.corner_radius = 25.0f;
	dock->diff_l = 0.0f;
	dock->diff_r = 0.0f;
	dock->flags = 0;
	dock->shrinked_allocation = (KibaAllocation) { 0.0f, 0.0f, 1.0f, 1.0f };
	
	priv->object_data = NULL;
	priv->dock_objects = NULL;
	priv->reposition = TRUE;
	priv->queue_update_surface = TRUE;
	priv->fade_alpha = 1.0f;
	priv->scroll_counter = 0;
	priv->dragging_child = NULL;
	priv->dragging_window = NULL;
	priv->press_count = 0;
	priv->drag_button = 0;
	priv->scroll_offset = 0.0f;
	
	// enabling drag and drop
	gtk_drag_dest_set (GTK_WIDGET (dock),
			GTK_DEST_DEFAULT_ALL,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
	gtk_drag_dest_add_uri_targets (GTK_WIDGET (dock));
}

/**
 * kiba_dock_new:
 * @configFile: configuration file associated to dock.
 *
 * Creates and returns a new KibaDock Container.
 *
 * Returns: The KibaDock Container created.
 */
GtkWidget*
kiba_dock_new (KibaPrefsFile *configFile) {
	GtkWidget *dock = g_object_new (KIBA_TYPE_DOCK,
		"config_file", configFile,
		NULL
	);
	gint i;
	
	// add preference callbacks
	kiba_prefs_add_options (configFile,
		dock_options, G_N_ELEMENTS(dock_options));
	for (i = 0; i < G_N_ELEMENTS(dock_options); i++) {
		kiba_prefs_add_callback (configFile, dock_options[i].id,
			preferences_changed_cb, dock, TRUE);
	}
	
	return dock;
}
