/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <librsvg/rsvg.h>
#include <cairo.h>
#include <math.h>

#include "config.h"

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "continuous-progress.h"

#define PROGRESS_ICON  KIBA_ICON_PATH"/gear.svg"
#define ANIMATE_RATE 65
#define ANIMATE_STEP 0.03
#define PROGRESS_DATA_KEY "continuous-progress"

G_DEFINE_TYPE (KibaContinuousProgressFeature,
	kiba_continuous_progress_feature, KIBA_TYPE_VISUAL_FEATURE)

/// Private methods

static gint
__animate (gpointer data)
{
	KibaContinuousProgressFeatureObject *cobject = data;
	KibaContinuousProgressFeature *progress = cobject->progress;
	KibaVisualFeatureObject *object = cobject->object;
	
	cobject->animate_value = MIN(2.0f,
		cobject->animate_value + ANIMATE_STEP);
	if (cobject->animate_value == 2.0f) {
		cobject->animate_value = 0.0f;
	}
	
	kiba_visual_feature_queue_draw(KIBA_VISUAL_FEATURE(progress), object);
	
	if (cobject->running) {
		if (cobject->alpha < 1.0f) {
			cobject->alpha += 0.05f;
		}
	} else {
		if (cobject->alpha > 0.0f) {
			cobject->alpha -= 0.05f;
		}
		if (cobject->alpha <= 0.0f) {
			kiba_visual_feature_remove_object(
				KIBA_VISUAL_FEATURE(progress), object->object);
			return FALSE;
		}
	}
	return TRUE;
}

static void
resize_feature_object(
		KibaVisualFeature *feature,
		KibaObject *object)
{
	
	GList *l;
	KibaVisualFeatureObject *vobject = NULL;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	if (vobject) {
		kiba_visual_feature_queue_resize(feature, vobject);
	}
}

static void
__scale_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__offset_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__working_toggled(KibaObject *object,
	gboolean active,
	gpointer data)
{
	KibaContinuousProgressFeatureObject *cobject;
	if (active) {
		cobject = g_new0(KibaContinuousProgressFeatureObject, 1);
		g_object_set_data(G_OBJECT(object),
			PROGRESS_DATA_KEY, cobject);
		
		cobject->progress = KIBA_CONTINUOUS_PROGRESS_FEATURE(data);
		cobject->alpha = 0.0f;
		cobject->animate_value = 0.0f;
		cobject->running = TRUE;
		cobject->timeout_id = g_timeout_add(ANIMATE_RATE, __animate, cobject);
		cobject->object = kiba_visual_feature_add_object(
			KIBA_VISUAL_FEATURE(data), object);
	} else {
		cobject = g_object_get_data(G_OBJECT(object), PROGRESS_DATA_KEY);
		if (cobject) { // feature maybe activated after the running state was set
			cobject->running = FALSE;
		}
	}
}

static void __object_size_changed (KibaDock *dock,
		gint new_size,
		gpointer data)
{
	KibaContinuousProgressFeature *progress = data;
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(progress);
	GList *l;
	
	progress->size = new_size*0.625f;
	
	kiba_visual_feature_queue_resize_all(feature);
}

/// Class implementation

static void
kiba_continuous_progress_feature_object_added(
		KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object))
		return;
	
	g_signal_connect(G_OBJECT(object), "working-toggled",
		G_CALLBACK(__working_toggled), feature);
	g_signal_connect(G_OBJECT(object), "scale-changed",
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_connect(G_OBJECT(object), "offset-changed",
		G_CALLBACK(__offset_changed_cb), feature);
	if (kiba_object_get_state(object, KIBA_OBJECT_ACTIVE)) {
		__working_toggled(object, TRUE, feature);
	}
}

static void
kiba_continuous_progress_feature_object_removed(
		KibaVisualFeature *feature, KibaObject *object)
{
	KibaContinuousProgressFeatureObject *cobject =
		g_object_get_data(G_OBJECT(object), PROGRESS_DATA_KEY);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__working_toggled), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__offset_changed_cb), feature);
	if (cobject) {
		if (cobject->timeout_id)
			g_source_remove(cobject->timeout_id);
		g_free(cobject);
	}
}

static void
kiba_continuous_progress_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaContinuousProgressFeature *progress =
		KIBA_CONTINUOUS_PROGRESS_FEATURE(feature);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaContinuousProgressFeatureObject *cobject =
		g_object_get_data(G_OBJECT(object->object), PROGRESS_DATA_KEY);
	cairo_surface_t *surface = NULL;
	gint surface_width = 0, surface_height = 0;
	gfloat alpha = cobject->alpha*
		KIBA_DRAWABLE_OBJECT(object->object)->alpha;
	
	if (alpha <= 0.0f) {
		return;
	}
	
	kiba_cached_icon_get_surface(
		progress->cached_icon,
		MAX(object->object->alloc.width,
		    object->object->alloc.height),
		&surface, &surface_width, &surface_height);
	
	cairo_scale(cr,
		object->allocation.width/surface_width,
		object->allocation.height/surface_height);
	
	cairo_translate(cr,
		surface_width/2.0f,
		surface_height/2.0f);
	cairo_rotate(cr, cobject->animate_value*M_PI);
	cairo_translate(cr,
		-surface_width/2.0f,
		-surface_height/2.0f);
	
	cairo_set_source_surface(cr, surface, 0, 0);
	cairo_paint_with_alpha(cr, alpha);
}

static KibaAllocation
kiba_continuous_progress_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	GdkRectangle area = kiba_drawable_object_get_area (drawable_object);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaContinuousProgressFeature *progress =
		KIBA_CONTINUOUS_PROGRESS_FEATURE(feature);
	KibaAllocation alloc;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];

	alloc = (KibaAllocation) {
		0.0f, 0.0f,
		progress->size,
		progress->size
	};
	
	alloc.x = ((float)d->halignment-1.0f)*object->object->surface_offset_x;
	alloc.y = ((float)d->valignment-1.0f)*object->object->surface_offset_y;
	
	alloc.width *= object->object->widget_scale_x;
	alloc.height *= object->object->widget_scale_y;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		alloc.y += object->object->alloc.height/2.0f + ((d->valignment-2)/2)*alloc.height;
		alloc.x += object->object->alloc.width/2.0f - alloc.width;
	} else {
		alloc.x += object->object->alloc.width/2.0f + ((d->halignment-2)/2)*alloc.width;
		alloc.y += object->object->alloc.height/2.0f - alloc.height;
	}
	
	return alloc;
}

static void kiba_continuous_progress_feature_get_max_offset(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		gfloat *left, gfloat *right,
		gfloat *top, gfloat *bottom)
{
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaContinuousProgressFeature *progress =
		KIBA_CONTINUOUS_PROGRESS_FEATURE(feature);
	const KibaLayoutData *d;
	gfloat widget_scale_x = 1.0f, widget_scale_y = 1.0f;
	GList *l;
	
	d = &kiba_layout_data[dock->d.position];
		
	for (l = object->object->scale_data; l != NULL; l = l->next) {
		KibaObjectScaleData *data = l->data;
		gfloat custom_factor_x = 1.0f;
		gfloat custom_factor_y = 1.0f;

		if (data->scale_widget) {
			data->max_func (object->object,
				&custom_factor_x,
				&custom_factor_y,
				data->data);
		
			widget_scale_x *= custom_factor_x;
			widget_scale_y *= custom_factor_y;
		}
	}
	widget_scale_x *= object->object->cfg_scale_x;
	widget_scale_y *= object->object->cfg_scale_y;
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*left = 0.0f;
		*top = progress->size*widget_scale_y // max feature size
		     -(object->object->size.height*widget_scale_y)/2.0f; // half max object size
	} else {
		*left = progress->size*widget_scale_x // max feature size
		     -(object->object->size.width*widget_scale_x)/2.0f; // half max object size
		*top = 0.0f;
	}
	*right = 0.0f;
	*bottom = 0.0f;
}

static void
kiba_continuous_progress_feature_win_set(
		KibaVisualFeature *feature, KibaWin *win)
{
	KibaContinuousProgressFeature *progress =
		KIBA_CONTINUOUS_PROGRESS_FEATURE(feature);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	cairo_t *cr;
	
	cr = kiba_win_cairo_create(win);
	progress->cached_icon = kiba_cached_icon_new(
		PROGRESS_ICON,
		cr);
	cairo_destroy(cr);
	
	__object_size_changed(dock, dock->d.object_size, progress);
	g_signal_connect(G_OBJECT(dock), "object-size-changed",
		G_CALLBACK(__object_size_changed), progress);
}

static void
kiba_continuous_progress_feature_dispose(GObject *object)
{
	KibaContinuousProgressFeature *progress =
		KIBA_CONTINUOUS_PROGRESS_FEATURE(object);
	if (KIBA_IS_CACHED_ICON(progress->cached_icon)) {
		g_object_unref(progress->cached_icon);
	}
}

static void
kiba_continuous_progress_feature_class_init (
		KibaContinuousProgressFeatureClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	// GObject overrides
	gobject_class->dispose = kiba_continuous_progress_feature_dispose;
}

static void
kiba_continuous_progress_feature_init (
		KibaContinuousProgressFeature *progress)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(progress);
	
	progress->cached_icon = NULL;
	progress->size = 0.0f;
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_continuous_progress_feature_get_allocation;
	feature->draw = kiba_continuous_progress_feature_draw;
	feature->object_added = kiba_continuous_progress_feature_object_added;
	feature->object_removed = kiba_continuous_progress_feature_object_removed;
	feature->win_set = kiba_continuous_progress_feature_win_set;
	feature->get_max_offset = kiba_continuous_progress_feature_get_max_offset;
}

KibaVisualFeature*
kiba_continuous_progress_feature_new(KibaWin *win)
{
	KibaVisualFeature *feature = g_object_new (
		    KIBA_TYPE_CONTINUOUS_PROGRESS_FEATURE,
		    "name",  CONTINUOUS_PROGRESS_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_GIMMICK,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	return feature;
}
