/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <cairo.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "greyed.h"

#define GREYED_OBJECT_KEY "greyed-object"

G_DEFINE_TYPE (KibaGreyedFeature, kiba_greyed_feature, KIBA_TYPE_VISUAL_FEATURE)

typedef struct {
	KibaVisualFeature *feature;
	KibaObject *object;
	gboolean grey_in;
	gint animate_id;
	gfloat alpha;
}GreyedObject;

/// Private methods

static gboolean
__animate (gpointer data)
{
	GreyedObject *gobject = data;
	KibaObject *object = gobject->object;
	
	kiba_drawable_object_queue_render_object(KIBA_DRAWABLE_OBJECT(object));
	
	if (gobject->grey_in) {
		gobject->alpha = MIN(1.0f, gobject->alpha+0.1f);
		
		if (gobject->alpha == 1.0f) {
			gobject->animate_id = 0;
			return FALSE;
		} else {
			return TRUE;
		}
	} else {
		gobject->alpha = MAX(0.0f, gobject->alpha-0.1f);
		
		if (gobject->alpha == 0.0f) {
			// remove greyed feature for object
			kiba_visual_feature_remove_object(gobject->feature, object);
			
			g_free(gobject);
			g_object_set_data(G_OBJECT(object), GREYED_OBJECT_KEY, NULL);
			
			return FALSE;
		} else {
			return TRUE;
		}
	}
}

static void
__greyed_toggled(KibaObject *object,
	gboolean greyed,
	gpointer data)
{
	GreyedObject *gobject = g_object_get_data(
		G_OBJECT(object),
		GREYED_OBJECT_KEY);
	
	if (greyed) {
		if (!gobject) {
			gobject = g_new0(GreyedObject, 1);
			gobject->alpha = 0.0f;
			gobject->object = object;
			gobject->feature = data;
			
			g_object_set_data(G_OBJECT(object),
				GREYED_OBJECT_KEY, gobject);
			
			kiba_visual_feature_add_object(KIBA_VISUAL_FEATURE(data), object);
		}
		gobject->grey_in = TRUE;
		
		if (!gobject->animate_id)
			gobject->animate_id = g_timeout_add(20, __animate, gobject);
	} else {
		if (!gobject) return;
		gobject->grey_in = FALSE;
		
		if (!gobject->animate_id)
			gobject->animate_id = g_timeout_add(20, __animate, gobject);
	}
}

/// Class implementation

static void
kiba_greyed_feature_object_added(KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object) || KIBA_IS_SEPARATOR(object)) {
		return;
	}
	
	g_signal_connect(G_OBJECT(object), "greyed-toggled",
		G_CALLBACK(__greyed_toggled), feature);
	if (kiba_object_get_state(object, KIBA_OBJECT_GREYED)) {
		kiba_visual_feature_add_object(feature, object);
	}
}

static void
kiba_greyed_feature_object_removed(KibaVisualFeature *feature,
		KibaObject *object)
{
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__greyed_toggled), feature);
	if (!kiba_object_get_state(object, KIBA_OBJECT_GREYED)) {
		kiba_visual_feature_remove_object(feature, object);
	}
}

static void
kiba_greyed_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaGreyedFeature *greyed = KIBA_GREYED_FEATURE(feature);
	GreyedObject *gobject = g_object_get_data(
		G_OBJECT(object->object),
		GREYED_OBJECT_KEY);
	if (gobject) {
		cairo_set_source_rgba (cr,
			0.2f,
			0.2f,
			0.2f,
			0.6f*gobject->alpha);
		cairo_paint (cr);
	}
}

static KibaAllocation
kiba_greyed_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	GdkRectangle area = kiba_drawable_object_get_area (
		KIBA_DRAWABLE_OBJECT(object->object));
	object->allocation =(KibaAllocation) {
		0.0f, 0.0f,
		area.width,
		area.height
	};
	return object->allocation;
}

static void
kiba_greyed_feature_class_init (KibaGreyedFeatureClass *class)
{
}

static void
kiba_greyed_feature_init (KibaGreyedFeature *greyed)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(greyed);
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_greyed_feature_get_allocation;
	feature->draw = kiba_greyed_feature_draw;
	feature->object_added = kiba_greyed_feature_object_added;
	feature->object_removed = kiba_greyed_feature_object_removed;
}

KibaVisualFeature*
kiba_greyed_feature_new(KibaWin *win)
{
	KibaVisualFeature *feature = g_object_new (KIBA_TYPE_GREYED_FEATURE,
		    "name",  GREYED_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_ONTOP,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	return feature;
}
