/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "kiba.h"
#include "kiba-private.h"
#include "kiba-dock.h"
#include "kiba-dialog.h"
#include "kiba-screen.h"
#include "kiba-plugin-loader.h"
#include "kiba-separator.h"
#include "kiba-tooltip.h"
#include "kiba-title.h"
#include "kiba-debug.h"
#include "kiba-utils.h"
#include "kiba-icon.h"
#include "kiba-container-object.h"

#include "kiba-private.h"
#include "prefs-private.h"
#include "dock-menu.h"

#ifdef USE_KDE
	const char *file_manager_global = "dolphin";
#else
#ifdef USE_XFCE
	const char *file_manager_global = "thunar";
#else
	const char *file_manager_global = "nautilus --no-desktop";
#endif
#endif

#define KIBA_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE, KibaPrivate))

#define _(string)  gettext (string)

struct _Kiba {
	GObject parentObject;

	KibaTitle  *title; /**< the KibaTitle instance, to display a title above objects */
	KibaTooltip *tooltipWin;
};

struct _KibaClass {
	GObjectClass  parentClass;
};

typedef struct {
	Kiba *kiba;
	KibaWin *win;
	GList *plugins; /**< list of KibaInternPluginData's */
	KibaPrefsFile *prefsFile;
}KibaInternDockData;

typedef struct {
	KibaPlugin *plugin;
	KibaPrefsFile *prefsFile;
	PrefMember *enableOption;
	KibaInternDockData *data; /* NULL for global plugins */
	KibaWin *win; /* NULL for global plugins */
}KibaInternPluginData;

typedef struct {
	GList *generalPlugins; /* list of global KibaInternPluginData's */
	GList *docks; /* list of KibaInternDockData's */

	gint move_dock_id; /* id for the dock drag timeout */
	GdkRectangle mouse_anchor;
}KibaPrivate;

G_DEFINE_TYPE (Kiba, kiba, G_TYPE_OBJECT)

static Kiba *kiba_instance = NULL;

/// Private Methods

static KibaInternDockData*
__get_dock_data_for_plugin(Kiba *kiba,
		KibaPlugin *plugin)
{
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	GList *l;

	for (l = priv->docks; l != NULL; l = l->next) {
		KibaInternDockData *data = l->data;
		GList *plugins;
		for (plugins = data->plugins; plugins != NULL; plugins = plugins->next) {
			KibaInternPluginData *pluginData = plugins->data;
			if (pluginData->plugin == plugin)
				return data;
		}
	}

	return NULL;
}

static KibaInternDockData* __get_data_for_win(Kiba *kiba,
                                        KibaWin *win) {
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	GList *l;

	for (l = priv->docks; l != NULL; l = l->next) {
		KibaInternDockData *data = l->data;
		if (data->win == win)
			return data;
	}

	return NULL;
}

/// signal handler

static gboolean
__dock_button_press_event (GtkWidget *widget,
		GdkEventButton *event,
		gpointer user_data)
{
	if (event->button == 3) {
		return kiba_dock_show_menu (widget, event);
	}
	return FALSE;
}

#if GTK_CHECK_VERSION(2,12,0) == TRUE
static gboolean
__object_query_tooltip (GtkWidget *widget,
		gint x,
		gint y,
		gboolean keyboard_mode,
		GtkTooltip *tooltip,
		gpointer user_data)
{
	Kiba *kiba = user_data;
	gchar *text = gtk_widget_get_tooltip_text(widget);
	if (!text) return FALSE;
	kiba_tooltip_set_text(kiba->tooltipWin, text);
	g_free(text);
	return TRUE;
}
#endif

static void
__dock_add_cb (GtkWidget *widget,
		GtkWidget *child,
		gpointer data)
{
#if GTK_CHECK_VERSION(2,12,0) == TRUE
	// use custom window for tooltip queries
	//gtk_widget_set_tooltip_window (widget, GTK_WINDOW(kiba_get()->tooltipWin));
	//g_signal_connect(G_OBJECT(widget), "query-tooltip",
	//		G_CALLBACK(__object_query_tooltip), data);
#endif
}

static KibaWin*
__create_dock (KibaPrefsFile *file)
{
	GtkWidget *window, *dock;
	GSList *l;

	// initiall the window
	window = kiba_win_new (file);
	// initiall dock
	dock = kiba_dock_new (file);
	gtk_container_add (GTK_CONTAINER (window), dock);
	DEBUG_INFO("KibaDock created.\n");

	g_signal_connect(G_OBJECT(dock), "add",
		G_CALLBACK(__dock_add_cb), NULL);
	g_signal_connect(G_OBJECT(dock), "button-press-event",
		G_CALLBACK(__dock_button_press_event), NULL);

	const GSList* dock_objects = kiba_prefs_get_list (file, "dock_objects", NULL);
	kiba_dock_set_object_list (KIBA_DOCK(dock), dock_objects);

	gtk_widget_show_all (GTK_WIDGET(window));

	return KIBA_WIN(window);
}

static void
__plugin_toggled_cb (const gchar *option_id,
		gpointer value,
		gpointer userData)
{
	Kiba *kiba = kiba_get();
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaInternPluginData *pluginData = userData;
	KibaPluginLoader *loader = kiba_plugin_loader_get();

	if ((gboolean)(GPOINTER_TO_INT (value))) {
		if (pluginData->plugin->loaded)
			return;
		kiba_plugin_loader_load_plugin (loader, pluginData->plugin);
	} else if (!(gboolean)(GPOINTER_TO_INT (value)) && pluginData->plugin) {
		GList *plugins, *l;
		GList *dependingPlugins = NULL;

		if (!pluginData->plugin->loaded)
			return;

		/* check if there is a loaded plugin depending on this plugin,
		   in this case we cant unload this plugin immediately */
		plugins = kiba_plugin_loader_get_plugins(loader);
		for (l = plugins; l != NULL; l = l->next) {
			KibaPlugin *plugin = l->data;
			KibaWin *plugWin = kiba_get_plugin_win(kiba, plugin);
			gint i;

			if (plugin == pluginData->plugin || !plugin->loaded)
				continue;

			for (i = 0; i < plugin->num_dependencies; i++) {
				KibaPluginDependency dependency = plugin->dependencies[i];

				if (dependency.type == KIBA_PLUGIN_DOCK && pluginData->win != plugWin)
					continue;

				if (strcmp(dependency.id, G_TYPE_MODULE(pluginData->plugin)->name) == 0) {
					KibaPluginDependencyData *depData = g_new0(KibaPluginDependencyData, 1);
					depData->id = g_strdup(G_TYPE_MODULE(plugin)->name);
					depData->type = kiba_plugin_type(plugin);
					depData->file = kiba_get_plugin_config_file(kiba, plugin);
					dependingPlugins = g_list_append(dependingPlugins, depData);
				}
			}
		}

		if (dependingPlugins) {
			kiba_prefs_set_option(pluginData->prefsFile, option_id, "bool", "TRUE");
			kiba_plugin_loader_popup_depend_on_dialog(pluginData->plugin, dependingPlugins, pluginData->prefsFile);
		} else {
			kiba_plugin_loader_unload_plugin(loader, pluginData->plugin);
		}
	}
}

static gboolean
__add_plugin (KibaInternPluginData *pluginData,
		KibaPrefsFile *prefsFile,
		const gchar *pluginPath)
{
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	gchar *pluginID, *buf;
	gchar *enableKey;

	pluginData->plugin = kiba_plugin_loader_new_plugin(loader, pluginPath);
	if (!pluginData->plugin)
		return FALSE;

	static PrefMember plugin_enable_options[] = {
		{ NULL, "FALSE", KIBA_BOOLEAN }
	};

	pluginData->prefsFile = prefsFile;

	buf = g_strrstr (pluginPath, "/") + 4;
	pluginID = g_strndup(buf, strlen(buf) - 3); /* 3 = strlen(".so") */

	enableKey = g_strdup_printf ("%s_enable", pluginID);
	plugin_enable_options->id = enableKey;
	pluginData->enableOption = plugin_enable_options;
	g_free(pluginID);
	kiba_prefs_add_options (prefsFile, plugin_enable_options, G_N_ELEMENTS(plugin_enable_options));
	kiba_prefs_add_callback (prefsFile, enableKey, __plugin_toggled_cb, pluginData, TRUE);

	return TRUE;
}

/*** methods that are only available to the core ***/

KibaPlugin*
kiba_get_plugin_for_id (Kiba *kiba,
		KibaPrefsFile *file,
		const gchar *id)
{
	GList *plugins;
	KibaPluginLoader *loader;
	GList *l;

	loader = kiba_plugin_loader_get();
	plugins = kiba_plugin_loader_get_plugins (loader);

	for (l = plugins; l != NULL; l = l->next) {
		KibaPlugin *plugin = l->data;
		GTypeModule *module = G_TYPE_MODULE(plugin);
		KibaPrefsFile *plugFile = kiba_get_plugin_config_file (kiba, plugin);

		if (plugFile != file)
			continue;

		if (strcmp (module->name, id) == 0) {
			return plugin;
		}
	}
	return NULL;
}

gboolean
kiba_add_dock_plugin (Kiba *kiba,
		KibaWin *win,
		const gchar *pluginPath)
{
	KibaInternPluginData *pluginData;
	KibaInternDockData *data;
	gboolean success;
	GList *l;

	data =  __get_data_for_win(kiba, win);
	if (!data) return FALSE;

	for (l = data->plugins; l != NULL; l = l->next) {
		KibaInternPluginData *pluginData = l->data;
		if (pluginData->prefsFile == data->prefsFile &&
		    strcmp(kiba_plugin_get_lib_path(pluginData->plugin), pluginPath) == 0)
			return pluginData->plugin ? pluginData->plugin->loaded : FALSE;
	}

	pluginData = g_new0(KibaInternPluginData, 1);
	pluginData->data = data;
	pluginData->win = data->win;

	data->plugins = g_list_append(data->plugins, pluginData);

	success = __add_plugin(pluginData, data->prefsFile, pluginPath);
	if (!success) {
		data->plugins = g_list_remove(data->plugins, pluginData);
		g_free(pluginData);
	}

	return success;
}

gboolean
kiba_add_general_plugin (const char *pluginPath)
{
	Kiba *kiba = kiba_get();
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaInternPluginData *pluginData;
	KibaPrefsFile *generalFile;
	gboolean success;

	generalFile = kiba_prefs_get_general_file(kiba_prefs_get());

	pluginData = g_new0(KibaInternPluginData, 1);
	pluginData->data = NULL;
	pluginData->win = NULL;

	success = __add_plugin(pluginData, generalFile, pluginPath);
	if (!success) {
		g_free(pluginData);
	} else {
		priv->generalPlugins = g_list_append(priv->generalPlugins, pluginData);
	}

	return success;
}

/// Public Methods

/**
 * kiba_add_plugin_object:
 * @kiba: #Kiba instance.
 * @plugin: #KibaPlugin instance.
 * @object: #KibaObject to add.
 *
 * Add @object to the dock associated to @plugin.
 *
 * Returns: TRUE on success, otherwise FALSE is returned.
 */
gboolean
kiba_add_plugin_object (Kiba *kiba,
		KibaPlugin *plugin,
		KibaObject *object)
{
	g_return_val_if_fail(KIBA_IS(kiba), FALSE);
	g_return_val_if_fail(KIBA_IS_OBJECT(object), FALSE);
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), FALSE);
	
	KibaInternDockData *data;
	GtkWidget *dockWidget;

	data = __get_dock_data_for_plugin(kiba, plugin);
	if (!data) {
		DEBUG_ERROR("Failed to get dock for plugin '%s' (%p)\n",
				G_TYPE_MODULE(plugin)->name, plugin);
		return FALSE;
	}

	dockWidget = KIBA_WIN_CHILD(data->win);
	gtk_container_add (GTK_CONTAINER(dockWidget), GTK_WIDGET(object));

	return TRUE;
}

/**
 * kiba_get_plugin_config_file:
 * @kiba: #Kiba instance.
 * @plugin: #KibaPlugin instance.
 *
 * Returns path to the config file associated to @plugin.
 * Can be called after #kiba_add_object method above was called.
 *
 * Returns: Path to the config file, or NULL if no config file was associated to @plugin.
 */
KibaPrefsFile*
kiba_get_plugin_config_file (Kiba *kiba,
		KibaPlugin *plugin)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), NULL);

	if (!KIBA_PLUGIN_GLOBAL(plugin)) {
		KibaInternDockData *data;

		data = __get_dock_data_for_plugin(kiba, plugin);
		if (!data) {
			DEBUG_ERROR("Failed to get the dock for plugin '%s' (%p)\n",
					G_TYPE_MODULE(plugin)->name, plugin);
			return NULL;
		}

		return data->prefsFile;
	}
	return kiba_prefs_get_general_file(kiba_prefs_get());
}

/**
 * kiba_get_win_config_file:
 * @kiba: #Kiba instance.
 * @win: #KibaWin instance.
 *
 * Returns path to the config file associated to @win.
 *
 * Returns: Path to the config file or NULL if no file is associated to @win.
 */
KibaPrefsFile*
kiba_get_win_config_file (Kiba *kiba,
		KibaWin *win)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	g_return_val_if_fail(KIBA_IS_WIN(win), NULL);
	
	KibaInternDockData *data;

	data = __get_data_for_win(kiba, win);
	if (!data) {
		DEBUG_ERROR("Failed to get the data for win (%p)\n", win);
		return NULL;
	}

	return data->prefsFile;
}

/**
 * kiba_get_dock_config_file:
 * @kiba: #Kiba instance.
 * @dock: #KibaDock instance.
 *
 * Returns: Path to the config file or NULL if no file is associated to @dock.
 */
KibaPrefsFile*
kiba_get_dock_config_file (Kiba *kiba,
		KibaDock *dock)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	g_return_val_if_fail(KIBA_IS_DOCK(dock), NULL);
	
	return  kiba_get_win_config_file(kiba,
		KIBA_WIN(GTK_WIDGET(dock)->parent));
}


/**
 * kiba_get_plugin_win:
 * @kiba: #Kiba instance.
 * @plugin: #KibaPlugin instance.
 *
 * Returns window associated to @plugin.
 *
 * Returns: a #KibaWin instance, or NULL if no #KibaWin is associated to @plugin.
 */
KibaWin*
kiba_get_plugin_win (Kiba *kiba,
		KibaPlugin *plugin)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), NULL);
	
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaInternDockData *data;
	GList *l;

	if (KIBA_PLUGIN_GLOBAL(plugin)) {
		return NULL;
	}

	for (l = priv->docks; l != NULL; l = l->next) {
		GList *plugins;
		data = l->data;
		for (plugins = data->plugins; plugins != NULL; plugins = plugins->next) {
			KibaInternPluginData *plugData = plugins->data;
			if (plugData->plugin == plugin)
				return data->win;
		}
	}

	return NULL;
}

/**
 * kiba_get_win:
 * @kiba: #Kiba instance.
 *
 * Returns first window found.
 *
 * Returns: a #KibaWin instance, or NULL if no #KibaWin was initialled before.
 */
KibaWin*
kiba_get_win (Kiba *kiba)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);

	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaInternDockData *data;

	if (!priv->docks) return NULL;
	data = priv->docks->data;
	return data->win;
}

/**
 * kiba_get_title:
 * @kiba: #Kiba Instance.
 *
 * Returns the Title widget.
 *
 * Returns: #KibaTitle instance.
 */
KibaTitle*
kiba_get_title (Kiba *kiba)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	return KIBA_TITLE(kiba->title);
}

/**
 * kiba_get_tooltip:
 * @kiba: #Kiba Instance.
 *
 * Returns the Tooltip widget.
 *
 * Returns: #KibaTooltip instance.
 */
KibaTooltip*
kiba_get_tooltip (Kiba *kiba)
{
	g_return_val_if_fail(KIBA_IS(kiba), NULL);
	return KIBA_TOOLTIP(kiba->tooltipWin);
}

/**
 * kiba_get_docks:
 * @kiba: #Kiba Instance.
 *
 * Returns a newly allocated list of all loaded #KibaDock instances.
 *
 * Returns: List of #KibaDock instances.
 */
GList*
kiba_get_docks(Kiba *kiba)
{
	GList *ret = NULL;
	GList *l;

	g_return_val_if_fail(KIBA_IS(kiba), NULL);

	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);

	for (l = priv->docks; l != NULL; l = l->next) {
		KibaInternDockData *data = l->data;
		if (GTK_IS_BIN(data->win))
			ret = g_list_append(ret, KIBA_WIN_CHILD(data->win));
	}

	return ret;
}

void
kiba_add_config_file (Kiba *kiba,
		const char *configFile)
{
	KibaPrivate *priv;
	KibaPrefsFile *prefsFile;
	KibaInternDockData *data;
	GtkWidget *dock;
	KibaPluginLoader *loader;
	GList *l;

	g_return_if_fail (KIBA_IS(kiba));
	g_return_if_fail (configFile != NULL);

	priv = KIBA_GET_PRIVATE (kiba);
	loader = kiba_plugin_loader_get();

	prefsFile = kiba_prefs_get_file(configFile);
	if (prefsFile) {
		DEBUG_WARNING("Config file '%s' allready added\n", configFile);
		return;
	}

	// add the config file to the database
	prefsFile = kiba_prefs_add_config_file(configFile);
	if (!prefsFile) {
		DEBUG_ERROR("Failed to add config file '%s'\n", configFile);
		return;
	}
	// update the file from schema
	kiba_prefs_update_config_file(kiba_prefs_get(), prefsFile);

	data = g_new0(KibaInternDockData, 1);
	data->kiba = kiba;
	data->plugins = NULL;
	data->prefsFile = prefsFile;

	priv->docks = g_list_append(priv->docks, data);

	// create the dock
	data->win = __create_dock(prefsFile);
	dock = KIBA_WIN_CHILD(data->win);

	// load the plugins enabled in this config file
	GList *plugins = kiba_plugin_loader_get_plugin_paths(loader);
	for (l = plugins; l != NULL; l = l->next) {
		KibaPlugin *loaded_plugin;
		KibaPluginLoaderPluginData *pluginData = l->data;
		char *pluginID, *buf;

		// skipping global plugins
		if (pluginData->type == KIBA_PLUGIN_GLOBAL)
			continue;

		buf = g_strrstr (pluginData->path, "/") + 4;
		pluginID = g_strndup(buf, strlen(buf) - 3); // 3 = strlen(".so")

		loaded_plugin = kiba_get_plugin_for_id (kiba, prefsFile, pluginID);
		g_free(pluginID);
		if (loaded_plugin) continue;

		kiba_add_dock_plugin(kiba, data->win, pluginData->path);
	}

	// adding separator objects
	GList *dock_objects = kiba_dock_get_object_list(KIBA_DOCK(dock), FALSE);
	for (l = dock_objects; l != NULL; l = dock_objects) {
		GtkWidget *object;
		char *object_id = l->data;
		
		dock_objects = g_list_remove(dock_objects, l->data);
		
		if (!g_str_has_prefix (object_id, "separator")) {
			g_free(object_id);
			continue;
		}

		object = kiba_separator_new (object_id);
		g_free(object_id);
		gtk_container_add (GTK_CONTAINER(dock), object);
		gtk_widget_show (object);
	}
	g_list_free(dock_objects);
}

static KibaInternDockData*
__get_data_for_config_file (Kiba *kiba,
		const char *configFile)
{
	KibaPrivate *priv;
	GList *l;

	priv = KIBA_GET_PRIVATE (kiba);

	for (l = priv->docks; l != NULL; l = l->next) {
		KibaInternDockData *data = l->data;
		if (strcmp(configFile, kiba_prefs_get_file_path(data->prefsFile)) == 0)
			return data;
	}

	return NULL;
}

void
kiba_remove_config_file (Kiba *kiba,
		const char *configFile)
{
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPrivate *priv;
	GList *plugins;

	g_return_if_fail (KIBA_IS(kiba));
	g_return_if_fail (configFile != NULL);

	priv = KIBA_GET_PRIVATE (kiba);

	KibaInternDockData *data = __get_data_for_config_file(kiba, configFile);
	if (!data) return;

	for (plugins = data->plugins; plugins != NULL; plugins = data->plugins) {
		KibaInternPluginData *plugData = plugins->data;
		if (plugData->plugin)
			kiba_plugin_loader_unload_plugin (loader, plugData->plugin);
		g_free(plugData);

		data->plugins = g_list_remove(data->plugins, plugins->data);
	}

	DEBUG_MISC("Destroying dock ..\n");
	gtk_widget_destroy(GTK_WIDGET(data->win));

	priv->docks = g_list_remove(priv->docks, data);
	g_free(data);
}

/**
 * kiba_quit:
 * @kiba: #Kiba Instance.
 *
 * Hide the docks and call #gtk_main_quit.
 */
void
kiba_quit (Kiba *kiba)
{
	KibaPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS(kiba));
	
	priv = KIBA_GET_PRIVATE (kiba);
	
	GList *docks = g_list_copy(priv->docks);
	gboolean continueLoop = TRUE;
	while (continueLoop) {
		continueLoop = FALSE;
		for (l = docks; l != NULL; l = l->next) {
			KibaInternDockData *data = l->data;
			if (kiba_win_quit(GTK_WIDGET(data->win))) {
				continueLoop = TRUE;
			}
		}
		usleep (10000);
	}
	
	gtk_main_quit();
}

/// Class Init/Fini

static void
kiba_dispose (GObject *obj)
{
	Kiba *kiba = KIBA(obj);
	KibaPrefs *prefs = kiba_prefs_get();
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	GList *l;

	for (l = priv->generalPlugins; l != NULL; l = priv->generalPlugins) {
		KibaInternPluginData *data = l->data;
		priv->generalPlugins = g_list_remove(priv->generalPlugins, l->data);
		if (data->plugin->loaded) {
			kiba_plugin_loader_unload_plugin (loader, data->plugin);
		}
		kiba_prefs_remove_options (data->prefsFile, data->enableOption);
		g_free(data);
	}
	g_list_free (priv->generalPlugins);
	
	for (l = priv->docks; l != NULL; l = priv->docks) {
		KibaInternDockData *data = l->data;
		GtkWidget *dock = KIBA_WIN_CHILD(data->win);
		GList *plugins, *m;
		GList *dock_object_ids = NULL, *dock_objects;

		{ // saving some preferences
			kiba_prefs_write_to_config_file (data->prefsFile,
				"edge_position", "char",
				kiba_layout_data[KIBA_DOCK(dock)->d.position].name);

			gchar *offset = g_strdup_printf ("%i", KIBA_WIN(gtk_widget_get_parent(
					GTK_WIDGET(dock)))->offset);
			kiba_prefs_write_to_config_file (data->prefsFile, "dock_edge_offset", "int", offset);
			g_free (offset);
			
			gint separator_objects = 0;
			dock_objects = kiba_dock_get_objects(KIBA_DOCK(dock));
			for (m = dock_objects; m != NULL; m = dock_objects) {
				KibaObject *object = m->data;
				gchar *object_id = NULL;
				
				dock_objects = g_list_remove(dock_objects, object);
				
				if (KIBA_IS_SEPARATOR(object)) {
					object_id = g_strdup_printf(
						"separator%i\0", separator_objects);
				} else {
					g_object_get(object,
						"object_id", &object_id,
						NULL);
				}
				
				dock_object_ids = g_list_append(dock_object_ids, object_id);
			}
			g_list_free(dock_objects);
			
			gchar *dock_objects_string = kiba_prefs_list_to_string(dock_object_ids);
			kiba_prefs_write_to_config_file (data->prefsFile,
				"dock_objects", "list", dock_objects_string);
			g_free (dock_objects_string);
			
			for (m = dock_object_ids; m != NULL; m = dock_object_ids) {
				gchar *object_id = m->data;
				dock_object_ids = g_list_remove(dock_object_ids, object_id);
				g_free(object_id);
			}
			g_list_free(dock_object_ids);
		}

		for (plugins = data->plugins; plugins != NULL; plugins = data->plugins) {
			KibaInternPluginData *plugData = plugins->data;
			if (plugData->plugin->loaded)
				kiba_plugin_loader_unload_plugin (loader, plugData->plugin);
			kiba_prefs_remove_options (plugData->prefsFile, plugData->enableOption);
			g_free(plugData);

			data->plugins = g_list_remove(data->plugins, plugins->data);
		}

		priv->docks = g_list_remove(priv->docks, l->data);

		DEBUG_MISC("Destroying dock ..\n");
		gtk_widget_destroy (GTK_WIDGET(data->win));
		g_free(data);
	}
	g_list_free (priv->docks);

	gchar *monitor = g_strdup_printf ("%i",
		kiba_screen_get_monitor(KIBA_SCREEN_DEFAULT));
	kiba_prefs_write_to_config_file (
		kiba_prefs_get_general_file (kiba_prefs_get()),
		"monitor_number", "int", monitor);
	g_free (monitor);

	// Chain up to the parent class
	G_OBJECT_CLASS(kiba_parent_class)->dispose (obj);
	
	if (KIBA_IS_TOOLTIP(kiba->tooltipWin))
		gtk_widget_destroy(GTK_WIDGET(kiba->tooltipWin));
	if (KIBA_IS_TITLE(kiba->title))
		gtk_widget_destroy (GTK_WIDGET(kiba->title));
}

static void
kiba_class_init (KibaClass *class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (class);

	object_class->dispose = kiba_dispose;

	g_type_class_add_private (object_class, sizeof (KibaPrivate));
}

static void
kiba_init (Kiba *kiba)
{
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPrivate *priv = KIBA_GET_PRIVATE(kiba);
	KibaPrefs *prefs = kiba_prefs_get();
	gchar *configPath, *userPath;
	gchar *dockConfigPath;
	gchar *generalConfigPath, *globalConfigPath;
	GDir *configDir;
	GList *l;

	kiba_instance = kiba;

	priv->docks = NULL;
	priv->generalPlugins = NULL;

	// create kiba home directory
	userPath = g_strdup_printf ("%s/.kiba-dock", g_get_home_dir ());
	if (!g_file_test (userPath, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (userPath, 7664) != 0) {
			DEBUG_ERROR ("Failed to create user directory at '%s'!\n", userPath);
			g_free(userPath);
			exit(EXIT_FAILURE);
		}
	}

	// create config directory
	configPath = g_strdup_printf ("%s/config", userPath);
	g_free(userPath);
	if (!g_file_test (configPath, G_FILE_TEST_IS_DIR)) {
		if (g_file_test (configPath, G_FILE_TEST_IS_REGULAR)) {
			g_remove(configPath);
		}
		if (g_mkdir (configPath, 7664) != 0) {
			DEBUG_ERROR ("Failed to create config directory at '%s'!\n", configPath);
			g_free(configPath);
			exit(EXIT_FAILURE);
		}
	}

	// adding general config file
	generalConfigPath = g_strdup_printf ("%s/general.xml", configPath);
	KibaPrefsFile *generalFile = kiba_prefs_add_file_with_type(generalConfigPath,
			KIBA_PREFS_FILE_GENERAL);
	kiba_prefs_update_config_file(prefs, generalFile);
	g_free(generalConfigPath);
	
	// window for object tooltips
	kiba->tooltipWin = kiba_tooltip_new();
	kiba->title = kiba_title_new ();
	gtk_widget_realize(GTK_WIDGET(kiba->title));

	// create dock config directory
	dockConfigPath = g_strdup_printf ("%s/docks", configPath);
	g_free(configPath);
	if (!g_file_test (dockConfigPath, G_FILE_TEST_IS_DIR)) {
		if (g_file_test (dockConfigPath, G_FILE_TEST_EXISTS))
			g_remove(dockConfigPath);
		if (g_mkdir (dockConfigPath, 7664) != 0) {
			DEBUG_ERROR ("Failed to create dock config directory at '%s'!\n", dockConfigPath);
			g_free(dockConfigPath);
			exit(EXIT_FAILURE);
		}
	}

	// adding global config file
	globalConfigPath = g_strdup_printf ("%s/" KIBA_PREFS_GLOBAL_FILE, dockConfigPath);
	kiba_prefs_add_file_with_type(globalConfigPath, KIBA_PREFS_FILE_GLOBAL);
	g_free(globalConfigPath);

	// load the plugins enabled in the general config file
	GList *plugins = kiba_plugin_loader_get_plugin_paths(loader);
	for (l = plugins; l != NULL; l = l->next) {
		KibaPlugin *loaded_plugin;
		KibaPluginLoaderPluginData *pluginData = l->data;
		char *pluginID, *buf;

		// skipping dock plugins
		if (pluginData->type != KIBA_PLUGIN_GLOBAL) {
			continue;
		}

		buf = g_strrstr (pluginData->path, "/") + 4;
		pluginID = g_strndup(buf, strlen(buf) - 3); // 3 = strlen(".so")

		loaded_plugin = kiba_get_plugin_for_id (kiba, generalFile, pluginID);
		g_free(pluginID);
		if (loaded_plugin) continue;

		kiba_add_general_plugin(pluginData->path);
	}

	// open the config directory for reading
	configDir = g_dir_open (dockConfigPath, 0, NULL);
	if (configDir) {
		gint numberOfDocks = 0;
		while (TRUE) {
			const char *file = g_dir_read_name(configDir);
			if (!file) break;
			if (!g_str_has_suffix(file, ".xml") || strcmp(file, KIBA_PREFS_GLOBAL_FILE) == 0)
				continue;
			numberOfDocks++;
		}
		g_dir_rewind(configDir);

		// make sure at least one dock config file is there
		if (!numberOfDocks) {
			gchar *defaultFile = g_strdup_printf("%s/default_dock.xml", dockConfigPath);
			if (!g_file_test (defaultFile, G_FILE_TEST_EXISTS)) {
				g_file_set_contents (defaultFile, "<config></config>\0", -1, NULL);
			}
			g_free(defaultFile);
		}

		// get all config files from @configDir
		while (TRUE) {
			const char *file = g_dir_read_name(configDir);

			// quit loop if there are no more files
			if (!file) break;

			if (!g_str_has_suffix(file, ".xml") ||
			    strcmp(file, KIBA_PREFS_GLOBAL_FILE) == 0) {
				continue;
			}
			
			char *configFile = g_strdup_printf("%s/%s", dockConfigPath, file);
			kiba_add_config_file(kiba, configFile);
			g_free(configFile);
		}
		g_free(dockConfigPath);
		g_dir_close (configDir);
		
		for (l = priv->docks; l != NULL; l = l->next) {
			KibaInternDockData *d = l->data;
			GtkWidget *dock = GTK_BIN(d->win)->child;
			GList *m;
			for (m = gtk_container_get_children(GTK_CONTAINER(dock));
						m != NULL; m = m->next) {
				kiba_object_set_destroy_factor(KIBA_OBJECT(m->data), 1.0f);
				kiba_object_update_scale_factor(KIBA_OBJECT(m->data));
				kiba_object_do_update_scale_factor(KIBA_OBJECT(m->data));
			}
			kiba_dock_queue_update_surface (KIBA_DOCK(dock));
		}
	} else {
		DEBUG_ERROR ("Failed to open config directory at '%s'!\n", dockConfigPath);
		g_free(dockConfigPath);
	}

	DEBUG_INFO("Kiba initialled!\n");
}

static Kiba*
kiba_new()
{
	return g_object_new(KIBA_TYPE, NULL);
}

/**
 * kiba_get:
 *
 * Returns the #Kiba instance.
 * Don't call #g_object_unref on it, this is done only once on exiting.
 *
 * Returns: #Kiba instance.
 */
Kiba*
kiba_get()
{
	if (!kiba_instance) {
		kiba_instance = kiba_new();
	}
	return kiba_instance;
}
