/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <glib/gthread.h>

#include "kiba.h"
#include "kiba-plugin-loader.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "kiba-private.h"
#include "prefs-dbus.h"
#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

int
main (int argc, char *argv[])
{
	Kiba *kiba;
	KibaPluginLoader *loader;
#ifdef USE_LIBSN
	KibaSn *sn;
#endif
	gchar *pluginPath = NULL;
	gchar *dockPluginPath, *globalPluginPath;
	gint i;

#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, DATADIR"/locale");
	textdomain (PACKAGE);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
#endif
	
	// use decimal point instead of comma
	setlocale (LC_NUMERIC, "C");
	setlocale (LC_CTYPE, "");
	setlocale (LC_MESSAGES, "");

	// parsing command line parameters
	for (i = 0; i < argc; i++) {
		if (g_strrstr (argv[i], "--help") != NULL ||
		    g_strrstr (argv[i], "-h") != NULL) {
			fprintf (stderr, "Usage: kiba-dock [--verbose] [--plugin-path plugin_path] ...\n");
			return EXIT_SUCCESS;
		} else if (g_strrstr (argv[i], "--verbose") != NULL ||
		         g_strrstr (argv[i], "-v") != NULL) {
			// enables debugging
			kiba_debug_init ();
		} else if (i < argc + 1) {
			if (g_strrstr (argv[i], "--plugin-path") != NULL) {
				pluginPath = argv[i+1];
				if (!g_file_test (pluginPath, G_FILE_TEST_EXISTS)) {
					fprintf (stderr, "Invalid Directory in --plugin-path argument: '%s'\n", pluginPath);
					pluginPath = NULL;
				}
			}
		}
	}
	
#ifdef USE_GLITZ
	DEBUG_INFO("Kiba was build with GLITZ support\n");
#else
	DEBUG_INFO("Kiba was build without GLITZ support\n");
	#ifdef HAVE_GLITZ
	#ifndef CAIRO_HAS_GLITZ_SURFACE
		DEBUG_MISC("Your current cairo installation does not have glitz Support.\n");
		DEBUG_MISC("Cairo does not include glitz support by default.\n"
			"To use glitz in kiba you must recompile cairo with '--enable-glitz'.\n");
	#endif
	#endif
#endif

	if (!pluginPath) {
		// default plugin path
		pluginPath = LIBDIR "/kiba-dock";
	}
	
	DEBUG_MISC("Initialling GUI toolkit...\n");
	gtk_disable_setlocale();
	g_thread_init(NULL);
	gdk_threads_init();
	gtk_init (&argc, &argv);
	// Must init display
	gdk_pixbuf_xlib_init(gdk_display, 0);

	DEBUG_MISC("Initialling SVG library..\n");
	rsvg_init ();

	// make the plugins at pluginPath loadable
	loader = kiba_plugin_loader_get();
	globalPluginPath = g_strdup_printf ("%s/global_plugins", pluginPath);
	if (!kiba_plugin_loader_initiall_plugins_at_path (
	    loader, KIBA_PLUGIN_GLOBAL, globalPluginPath))
	{
		// no global plugins installed
	}
	g_free(globalPluginPath);
	dockPluginPath = g_strdup_printf ("%s/dock_plugins", pluginPath);
	if (!kiba_plugin_loader_initiall_plugins_at_path (loader, KIBA_PLUGIN_DOCK, dockPluginPath)) {
		DEBUG_ERROR("No plugins found at '%s'.", dockPluginPath);
		g_object_unref (G_OBJECT(loader));
		g_free(dockPluginPath);
		return EXIT_FAILURE;
	}
	g_free(dockPluginPath);
	
	// set up the preferences database
	kiba_prefs_get ();
	// set up the dbus interface
	prefs_dbus_get();

#ifdef USE_LIBSN
	sn = kiba_sn_get ();
#endif
	// creating the docks 
	kiba = kiba_get();

	// show tips dialog
 	kiba_dialog_tips ();

	// enter the GTK mainloop
	gdk_threads_enter();
	gtk_main();
	gdk_threads_leave();

	// cleanup

	DEBUG_MISC("Destroying docks..\n");
	g_object_unref (G_OBJECT(kiba));

	DEBUG_MISC("Cleaning up..\n");
	g_object_unref (G_OBJECT(loader));
#ifdef USE_LIBSN
	g_object_unref(G_OBJECT(sn));
#endif
	prefs_dbus_fini ();
	kiba_screen_destroy (kiba_screen_get());
	kiba_prefs_fini ();
	kiba_icon_factory_destroy();
	DEBUG_MISC("Cleaned up!\n");

	rsvg_term ();

	return EXIT_SUCCESS;
}
