/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <math.h>

#include "kiba-dock.h"
#include "kiba-separator.h"
#include "kiba-private.h"

#define SMALL_BORDERWIDTH(width) (2*width)/3

static void __draw_podest_hat (KibaDock *dock,
						    cairo_t  *cr) {
	const KibaLayoutData *d;
	gint border_width;
	gint small_border_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));
	small_border_width = SMALL_BORDERWIDTH(border_width);
	
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		if (d->valignment == 2) { /* dock at the bottom edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* bottom-left point */
				2*small_border_width,
				dock->surface_data.height - border_width);
			cairo_line_to (cr, /* top-left point */
				small_border_width/2 + dock->shape.corner_radius,
				small_border_width/2);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - small_border_width/2 - dock->shape.corner_radius,
				small_border_width/2);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - 2*small_border_width,
				dock->surface_data.height - border_width);
			cairo_line_to (cr, /* bottom-left point */
				2*small_border_width,
				dock->surface_data.height - border_width);
		}
		else { /* dock at the top edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-left point */
				2*small_border_width,
				border_width);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - 2*small_border_width,
				border_width);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - small_border_width/2 - dock->shape.corner_radius,
				dock->surface_data.height);
			cairo_line_to (cr, /* bottom-left point */
				small_border_width/2 + dock->shape.corner_radius,
				dock->surface_data.height);
			cairo_line_to (cr, /* top-left point */
				2*small_border_width,
				border_width);
		}
	}
	else {
		if (d->halignment == 0) { /* dock at the left edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-left point */
				border_width,
				2*small_border_width);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width,
				small_border_width/2 + dock->shape.corner_radius);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width,
				dock->surface_data.height - small_border_width/2 - dock->shape.corner_radius);
			cairo_line_to (cr, /* bottom-left point */
				border_width,
				dock->surface_data.height - 2*small_border_width);
			cairo_line_to (cr, /* top-left point */
				border_width,
				2*small_border_width);
		}
		else {
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-right point */
				dock->surface_data.width - border_width,
				2*small_border_width);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - border_width,
				dock->surface_data.height - 2*small_border_width);
			cairo_line_to (cr, /* bottom-left point */
				0,
				dock->surface_data.height - small_border_width/2 - dock->shape.corner_radius);
			cairo_line_to (cr, /* top-left point */
				0,
				small_border_width/2 + dock->shape.corner_radius);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - border_width,
				2*small_border_width);
		}
	}
	
	cairo_close_path (cr);
}

static void __draw (KibaDock *dock,
					 cairo_t  *cr,
					 gpointer  data) {
	const KibaLayoutData *d;
	cairo_pattern_t  *pattern;
	gint border_width;
	gint small_border_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));
	small_border_width = SMALL_BORDERWIDTH(border_width);
	
	d = &kiba_layout_data[dock->d.position];
	
	cairo_save (cr);
		cairo_set_line_width (cr, border_width);
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		
		cairo_set_source_rgba (cr,
			dock->style.border_rgba.r, dock->style.border_rgba.g,
			dock->style.border_rgba.b, dock->style.border_rgba.a);
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			if (d->valignment == 2) { /* dock at the bottom edge */
				cairo_move_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* bottom-left point */
					small_border_width,
					dock->surface_data.height - small_border_width);
				cairo_line_to (cr, /* top-left point */
					small_border_width + dock->shape.corner_radius,
					small_border_width);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - small_border_width - dock->shape.corner_radius,
					small_border_width);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - small_border_width,
					dock->surface_data.height - small_border_width);
				cairo_stroke (cr);
			}
			else { /* dock at the top edge */
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2 + dock->shape.corner_radius,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2 - dock->shape.corner_radius,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_stroke (cr);
			}
		}
		else {
			if (d->halignment == 0) { /* dock at the left edge */
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - small_border_width,
					border_width/2 + dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - small_border_width,
					dock->surface_data.height - border_width/2 - dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
			}
			else {
				cairo_move_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-left point */
					small_border_width,
					border_width/2 + dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					small_border_width,
					dock->surface_data.height - border_width/2 - dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
			}
		}
	cairo_restore (cr);
	
	cairo_save (cr);
		__draw_podest_hat (dock, cr);
		
		pattern = kiba_dock_get_pattern (dock);
		cairo_set_source (cr, pattern);
		cairo_fill (cr);
		cairo_pattern_destroy (pattern);
	cairo_restore (cr);
}

static void __clip_func (KibaDock *dock,
					      cairo_t  *cr,
					      gpointer  data) {
	const KibaLayoutData *d;
	gint x, y, width, height;
	gint border_width;
	
	border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	
	x = GTK_WIDGET(dock)->allocation.x;
	y = GTK_WIDGET(dock)->allocation.y;
	width = GTK_WIDGET(dock)->allocation.width;
	height = GTK_WIDGET(dock)->allocation.height;
	
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		cairo_rectangle (cr,
			x, y - ((d->valignment-2)/2)*border_width,
			width, height - (d->valignment/2)*border_width);
	else
		cairo_rectangle (cr,
			x - ((d->halignment-2)/2)*border_width, y,
			width - (d->halignment/2)*border_width, height);
	cairo_clip (cr);
}

void kiba_dock_podest_shape_activate (KibaDock *dock) {
	dock->shape.draw_func = __draw;
	dock->shape.clip_func = __clip_func;
	dock->shape.draw_data = NULL;
	
	kiba_dock_queue_update_surface(dock);
}
