/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <string.h>

#include "kiba-icon-factory.h"
#include "kiba-app-chooser.h"
#include "kiba-utils.h"

#define _(string)  gettext (string)
#define WINDOW_ROLE "kiba-app-chooser"

G_DEFINE_TYPE (KibaAppChooser, kiba_app_chooser, G_TYPE_OBJECT)

enum {
	SIG_APP_ACTIVATE = 0,
	SIG_KIBA_APP_CHOOSER_LAST
};
static guint kiba_app_chooser_signals[SIG_KIBA_APP_CHOOSER_LAST] = { 0 };

static gchar *section_data[] = {
	"All", "System", "Settings", "Office", "Graphic", "Audio/Video",
	"Network", "Games", "Development", "Tools"
};

static KibaAppChooserSection
kiba_app_chooser_section_from_string (const gchar *section)
{
	gint i;
	for (i = 0; i < G_N_ELEMENTS(section_data); i++) {
		if (strcmp (section_data[i], section) == 0)
			return i;
	}
	return KIBA_APP_SECTION_ALL;
}

static void
destroy_cb (GtkWidget *widget,
		gpointer data)
{
	KibaAppChooser *chooser = data;
	KibaAppChooserApp *app;
	GList *l;
	
	if (chooser->timeout_id != 0) {
		g_source_remove(chooser->timeout_id);
		chooser->timeout_id = 0;
	}

	for (l = chooser->apps; l != NULL; l = chooser->apps) {
		app = l->data;

		if (GTK_IS_WIDGET(app->box)) {
			gtk_widget_destroy (app->box);
		}
		g_free (app->uri);
		if (app->icon) {
			g_free (app->icon);
		}
		chooser->apps = g_list_remove (chooser->apps, app);
		g_free (app);
	}
	g_list_free (chooser->apps);
	g_object_unref (chooser);
}

static gboolean
app_button_press_callback (GtkWidget *widget,
		GdkEventButton *event,
		gpointer user_data)
{
	KibaAppChooserApp *app = user_data;

	kiba_icon_view_select_child (KIBA_ICON_VIEW(
		app->chooser->view_win->icon_view), app->box);
	gtk_widget_queue_draw (app->chooser->view_win->icon_view);

	if (event->type == GDK_2BUTTON_PRESS) {
		g_signal_emit (app->chooser,
			kiba_app_chooser_signals[SIG_APP_ACTIVATE],
			0, app->uri);
		
		gtk_widget_destroy (app->chooser->view_win->win);
		
		return TRUE;
	}
	
	return FALSE;
}

static void
combobox_changed (GtkComboBox *combobox,
		gpointer user_data)
{
	KibaAppChooser *chooser = user_data;
	KibaAppChooserApp *app;
	KibaAppChooserSection selected_section;
	KibaIconView *icon_view;
	GList *l;
	
	icon_view = KIBA_ICON_VIEW(chooser->view_win->icon_view);
	
	selected_section = kiba_app_chooser_section_from_string (
				gtk_combo_box_get_active_text(combobox));
	
	for (l = chooser->apps; l != NULL; l = l->next) {
		app = l->data;
		
		if (!app->loaded) {
			continue;
		}
		
		if (selected_section == KIBA_APP_SECTION_ALL || app->section == selected_section) {
			if (!app->added) {
				kiba_icon_view_win_add (chooser->view_win, app->box);
				gtk_widget_show (app->box);
				app->added = TRUE;
			}
		} else if (app->added) {
			// icon view unrefs box
			app->box = gtk_widget_ref(app->box);
			app->image->surface = cairo_surface_reference (app->image->surface);
			
			kiba_icon_view_win_remove (chooser->view_win, app->box);
			gtk_widget_hide (app->box);
			app->added = FALSE;
		}
	}

	chooser->active_section = selected_section;
}

static void
icon_view_child_selected (KibaIconView *icon_view,
		GtkWidget *child,
		gpointer user_data)
{
	KibaAppChooser *chooser = user_data;
	KibaAppChooserApp *app = NULL;
	gchar *markup;
	GList *l;

	for (l = chooser->apps; l != NULL; l = l->next) {
		KibaAppChooserApp *check_app = l->data;

		if (check_app->box == child) {
			app = check_app;
			break;
		}
	}

	if (!app) {
		return;
	}
	
	kiba_icon_view_select_child (KIBA_ICON_VIEW(
		chooser->view_win->icon_view), app->box);
	markup = g_markup_printf_escaped (
		"<span size=\"large\" weight=\"bold\">%s</span>", app->name);
	gtk_label_set_markup (GTK_LABEL(chooser->name_label), markup);
	g_free (markup);
	gtk_label_set_text (GTK_LABEL(
		chooser->description_label), app->description);
	gtk_widget_queue_draw (chooser->view_win->icon_view);
}

static gboolean
kiba_icon_view_expose_cb (GtkWidget *widget,
		GdkEventExpose *expose,
		gpointer data)
{
	KibaAppChooser *chooser = data;
	KibaInfoWin *win;
	cairo_t *cr = NULL;
	KibaRGBA borderColor;

	if (!GTK_WIDGET_REALIZED (chooser->view_win->win)) {
		return FALSE;
	}

	win = KIBA_INFO_WIN(chooser->view_win->win);

	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS) {
		return FALSE;
	}

	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);

	borderColor = kiba_get_info_win_border_color();

	cairo_set_source_rgba (cr,
		borderColor.r,
		borderColor.g,
		borderColor.b,
		borderColor.a);
	cairo_set_line_width (cr, 3);

	cairo_move_to (cr,
		0,
		chooser->name_label->allocation.y - 5);
	cairo_rel_line_to (cr,
		chooser->view_win->win->allocation.width,
		0);
	cairo_stroke (cr);

	cairo_move_to (cr,
		0,
		chooser->filter->allocation.y + 5 +
		chooser->filter->allocation.height);
	cairo_rel_line_to (cr,
		chooser->view_win->win->allocation.width,
		0);
	cairo_stroke (cr);

	cairo_destroy (cr);

	return FALSE;
}

static void
drag_data_get (GtkWidget *widget,
		GdkDragContext *drag_context,
		GtkSelectionData *data,
		guint info,
		guint time,
		gpointer user_data)
{
	KibaAppChooserApp *app = user_data;
	gtk_selection_data_set(data,
		GDK_SELECTION_TYPE_STRING,
		8, // 8 bits per character.
		(guchar *)app->uri,
		strlen(app->uri));
}

static void
box_destroyed(GtkWidget *widget,
		gpointer data)
{
	if (data) {
		kiba_icon_factory_unload_icon(
			(KibaIconFactoryIconInfo*) data);
	}
}

static void
load_application_icon (GdkPixbuf *pixbuf,
		RsvgHandle *handle,
		gint size,
		gpointer data)
{
	KibaAppChooserApp *app = data;
	if (RSVG_IS_HANDLE(handle)) {
		kiba_image_set_svg_handle(app->image, handle, size, size);
	} else if (GDK_IS_PIXBUF(pixbuf)) {
		kiba_image_set_pixbuf(app->image, pixbuf, size, size);
	}
}

static void
add_apps_from_directory (KibaAppChooser *chooser,
		const gchar *app_path)
{
	static const gchar desktop_entry_group[] = "Desktop Entry";
	static const gchar desktop_file_suffix[] = ".desktop";
	static const struct {
	    const gchar *name;
	    KibaAppChooserSection section;
	} categories_data[] = {
		{ "AudioVideo", KIBA_APP_SECTION_AUDIO_VIDEO },
		{ "WebBrowser", KIBA_APP_SECTION_NETWORK },
		{ "Game", KIBA_APP_SECTION_GAMES },
		{ "Graphics", KIBA_APP_SECTION_GRAPHICS },
		{ "Office", KIBA_APP_SECTION_OFFICE },
		{ "Network", KIBA_APP_SECTION_NETWORK },
		{ "Development", KIBA_APP_SECTION_DEVELOPMENT },
		{ "Utility", KIBA_APP_SECTION_TOOLS },
		{ "Filemanager", KIBA_APP_SECTION_TOOLS },
		{ "Accessibility", KIBA_APP_SECTION_TOOLS },
		{ "Settings", KIBA_APP_SECTION_SETTINGS },
		{ "X-XFCE", KIBA_APP_SECTION_SYSTEM },
		{ "KDE", KIBA_APP_SECTION_SYSTEM },
		{ "System", KIBA_APP_SECTION_SYSTEM },
		{ "GNOME", KIBA_APP_SECTION_SYSTEM }
	};
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};

	KibaAppChooserApp *app;
	GtkWidget *label;
	GDir *app_dir;
	gchar *filename;
	gchar *markup;
	GError *error = NULL;
	gint i;

	app_dir = g_dir_open (app_path, 0, &error);
	g_clear_error (&error);
	error = NULL;

	if (!app_dir) {
		return;
	}

	while ((filename = (gchar*) g_dir_read_name (app_dir)) != NULL) {
		GtkWidget *vbox;
		GKeyFile *file;
		gchar *icon;
		gchar *path;
		gchar *name;
		gchar *categories;
		gchar *description;

		path = g_strdup_printf ("%s/%s", app_path, filename);

		if (g_file_test (path, (G_FILE_TEST_IS_DIR)) == TRUE) {
			add_apps_from_directory (chooser, path);
			g_free (path);
			continue;
		}
		if (!g_str_has_suffix (filename, desktop_file_suffix)) {
			g_free (path);
			continue;
		}

		file = g_key_file_new ();
		if (!g_key_file_load_from_file (file, path, 0, &error)) {
			g_key_file_free (file);
			g_clear_error (&error);
			g_free (path);
			continue;
		}

		icon = g_key_file_get_value (file, desktop_entry_group, "Icon", &error);
		g_clear_error (&error);
		error = NULL;
		if (!icon) {
			g_free (path);
			continue;
		}

		categories = g_key_file_get_value (file, desktop_entry_group, "Categories", &error);
		g_clear_error (&error);
		error = NULL;

		name = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Name",
				NULL,
				&error);
		g_clear_error (&error);
		error = NULL;
		if (!name) {
			name = g_key_file_get_value (file,
				desktop_entry_group, "Name", &error);
			g_clear_error (&error);
			error = NULL;
		}

		description = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Comment",
				NULL,
				&error);
		g_clear_error (&error);
		error = NULL;
		if (!description) {
			description = g_key_file_get_value (file,
				desktop_entry_group, "Comment", &error);
			g_clear_error (&error);
			error = NULL;
		}
		g_key_file_free (file);

		app = g_new0 (KibaAppChooserApp, 1);

		vbox = gtk_vbox_new(FALSE, 0);

		app->image = KIBA_IMAGE(kiba_image_new());
		// enabling drag and drop
		gtk_drag_source_set (GTK_WIDGET(app->image),
			GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_COPY);
		gtk_widget_set_usize (GTK_WIDGET(app->image), 48, 48);
		gtk_widget_show (GTK_WIDGET(app->image));
		gtk_box_pack_start (GTK_BOX (vbox),
			GTK_WIDGET(app->image), FALSE, FALSE, 10);

		app->section = -1;
		if (categories) {
			for (i = 0; i < G_N_ELEMENTS(categories_data); i++) {
				const gchar *section = categories_data[i].name;
				if (g_strrstr (categories, section)) {
					app->section = categories_data[i].section;
					break;
				}
			}
		}
		if (app->section == -1) {
			app->section = KIBA_APP_SECTION_TOOLS;
		}

		label = gtk_label_new(name);
		gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
		gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
		gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
		markup = g_markup_printf_escaped ("<span size=\"small\">%s</span>",
				GTK_LABEL (label)->text);
		gtk_label_set_markup (GTK_LABEL (label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
		gtk_widget_show(label);

		app->box = vbox;
		app->uri = path;
		app->icon = icon;
		app->loaded = FALSE;
		app->added = FALSE;
		app->chooser = chooser;
		app->name = name;
		app->description = description;
		
		chooser->apps = g_list_append (chooser->apps, app);
		
		g_signal_connect (G_OBJECT (vbox),
				  "button_press_event",
				  G_CALLBACK (app_button_press_callback),
				  app);
		g_signal_connect (GTK_OBJECT (app->image),
				  "drag-data-get",
				  G_CALLBACK(drag_data_get),
				  app);
	}
	g_dir_close (app_dir);
}

static gboolean
filter_matches(KibaAppChooser *chooser,
		KibaAppChooserApp *app)
{
	gchar *name = NULL;
	gchar *desc = NULL;
	const gchar *entry_text;
	gchar *filter_text;
	gboolean match;

	entry_text = gtk_entry_get_text(GTK_ENTRY(chooser->filter));
	if (!entry_text || strcmp(entry_text, "") == 0) {
		return TRUE;
	}
	filter_text = g_strdup(entry_text);
	kiba_strdown (&filter_text);
		
	if (app->name) {
		name = g_strdup(app->name);
		kiba_strdown (&name);
	}
	
	if (app->description) {
		desc = g_strdup(app->description);
		kiba_strdown (&desc);
	}
	
	match = (name && g_strrstr (name, filter_text)) ||
		(desc && g_strrstr (desc, filter_text));

	if (name) {
		g_free (name);
	}
	if (desc) {
		g_free (desc);
	}
	g_free(filter_text);
	
	return match;
}

static void
filter_activate_cb (GtkEntry *entry,
		gpointer  data)
{
	KibaAppChooser *chooser = data;
	KibaAppChooserApp *app;
	KibaIconView *icon_view;
	gint active_icons = 0;
	GList *l;

	icon_view = KIBA_ICON_VIEW(chooser->view_win->icon_view);
	
	// get filtered applications
	for (l = chooser->apps; l != NULL; l = l->next) {
		app = l->data;
		
		if (!app->loaded) {
			continue;
		}
		
		if (chooser->active_section != KIBA_APP_SECTION_ALL &&
		    app->section != chooser->active_section) {
			continue;
		}
		
		if (!app->added) {
			if (filter_matches(chooser, app)) {
				gtk_widget_show (app->box);
				kiba_icon_view_win_add (chooser->view_win, app->box);
				app->added = TRUE;
				active_icons++;
			}
		} else if (!filter_matches(chooser, app)) {
			// icon view unrefs box
			app->box = gtk_widget_ref(app->box);
			app->image->surface = cairo_surface_reference (app->image->surface);
			
			kiba_icon_view_win_remove (chooser->view_win, app->box);
			gtk_widget_hide (app->box);
			app->added = FALSE;
		}
	}

	if (active_icons == 0) {
		// no application selected
		
		gtk_label_set_text (GTK_LABEL(chooser->description_label), NULL);
		gtk_label_set_text (GTK_LABEL(chooser->name_label), NULL);
	} else {
		// show selected app description/name
		
		gchar *markup;
		
		for (l = chooser->apps; l != NULL; l = l->next) {
			app = l->data;
			
			if (kiba_icon_view_get_selected_child(icon_view) == app->box) {
				break;
			}
		}
		
		markup = g_markup_printf_escaped (
				"<span size=\"large\" weight=\"bold\">%s</span>",
				app->name);
		gtk_label_set_markup (GTK_LABEL(chooser->name_label), markup);
		g_free (markup);
		gtk_label_set_text (GTK_LABEL(
			chooser->description_label), app->description);
	}
}

static gint
process_apps(gpointer data)
{
	KibaAppChooser *app_chooser = data;
	GList *l;
	guint count = 0;
	
	if (!KIBA_IS_APP_CHOOSER(app_chooser)) {
		app_chooser->timeout_id = 0;
		return FALSE;
	}
	
	// TODO: add icons until a maximal time is reached
	do {
		KibaAppChooserApp *app = app_chooser->currently_processing->data;
		KibaIconFactoryIconInfo *icon_info;
		
		app->loaded = TRUE;
		
		if ((app_chooser->active_section == KIBA_APP_SECTION_ALL ||
		    app->section == app_chooser->active_section &&
		    filter_matches(app_chooser, app))) {
			kiba_icon_view_win_add (app_chooser->view_win, app->box);
			app->added = TRUE;
		}
		
		icon_info = kiba_icon_factory_load_icon (app->icon, 48,
			load_application_icon, app);
		g_signal_connect (G_OBJECT (app->box),
				  "destroy",
				  G_CALLBACK (box_destroyed),
				  icon_info);

		app_chooser->currently_processing =
			app_chooser->currently_processing->next;
		
		count++;
		if (count > 5) break;
	} while(app_chooser->currently_processing != NULL);
	
	if (app_chooser->currently_processing != NULL) {
		return TRUE;
	} else {
		app_chooser->timeout_id = 0;
		return FALSE;
	}
}

/// Class functions

static void
kiba_app_chooser_class_init (KibaAppChooserClass *class)
{
	/**
	 * KibaDock::app-activate:
	 * @dock: the object on which the signal is emitted
	 * @uri: application key file uri.
	 *
	 * The app-activate signal is emitted when a application has been activated.
	 */
	kiba_app_chooser_signals[SIG_APP_ACTIVATE] =
		g_signal_new ("app-activate",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaAppChooserClass, app_activate),
			NULL, NULL,
			g_cclosure_marshal_VOID__STRING,
			G_TYPE_NONE, 1,
			G_TYPE_STRING);
}

static void
kiba_app_chooser_init (KibaAppChooser *chooser)
{
	chooser->apps = NULL;
	chooser->view_win = NULL;
}

/**
 * kiba_app_chooser_new:
 * @app_path: The Path to .desktop Application file(s).
 *
 * Creates and returns a KibaAppChooser Instance.
 *
 * Returns: The KibaAppChooser Instance created.
 */
KibaAppChooser*
kiba_app_chooser_new (const gchar *app_path)
{
	KibaAppChooser *chooser;
	KibaAppChooserApp *app;
	GtkWidget *box;
	GtkWidget *section_label;
	GtkWidget *filter_label;
	GtkRequisition req;
	gchar *markup;
	GList *l;
	gint i;

	g_return_val_if_fail (app_path != NULL, NULL);

	chooser = g_object_new (KIBA_TYPE_APP_CHOOSER, NULL);
	chooser->apps = NULL;

	chooser->view_win = kiba_icon_view_win_new(
		KIBA_ICON_VIEW_BOX_POSITION_RIGHT, 6, 4, TRUE, TRUE, FALSE);
	gtk_window_set_role (GTK_WINDOW(chooser->view_win->win),
		WINDOW_ROLE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(
		chooser->view_win->win), TRUE);
	gtk_window_set_position (GTK_WINDOW(
		chooser->view_win->win), GTK_WIN_POS_CENTER);

	// create a tree model containing all sections
	box = gtk_hbox_new (FALSE, 5);
		section_label = gtk_label_new(NULL);
		gtk_misc_set_alignment (GTK_MISC(section_label), 0.1, 0.5);
		markup = g_markup_printf_escaped (
			"<span weight=\"bold\">%s</span>", _("Section"));
		gtk_label_set_markup (GTK_LABEL (section_label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (box),
			section_label, FALSE, FALSE, 5);
		gtk_widget_show(section_label);

		chooser->combobox = gtk_combo_box_new_text ();
		for (i = 0; i < KIBA_APP_SECTION_LAST; i++) {
			const gchar *section = section_data[i];
			gtk_combo_box_append_text (GTK_COMBO_BOX(
				chooser->combobox), section);
		}
		gtk_combo_box_set_active(GTK_COMBO_BOX(chooser->combobox), 0);
		gtk_widget_set_usize (chooser->combobox, 50, -1);
		gtk_box_pack_start (GTK_BOX (box),
			chooser->combobox, TRUE, TRUE, 5);
		gtk_widget_show (chooser->combobox);
		gtk_box_pack_start (GTK_BOX (chooser->view_win->top_box),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	// filter box
	box = gtk_hbox_new (FALSE, 5);
		filter_label = gtk_label_new(NULL);
		gtk_misc_set_alignment (GTK_MISC(filter_label), 0.1, 0.5);
		markup = g_markup_printf_escaped (
			"<span weight=\"bold\">%s</span>", _("Filter"));
		gtk_label_set_markup (GTK_LABEL (filter_label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (box),
			filter_label, FALSE, FALSE, 5);
		gtk_widget_show(filter_label);

		chooser->filter = gtk_entry_new();
		gtk_box_pack_start (GTK_BOX (box),
			chooser->filter, TRUE, TRUE, 5);
		gtk_widget_show(chooser->filter);
		g_signal_connect (G_OBJECT(chooser->filter), "activate",
			G_CALLBACK(filter_activate_cb), chooser);

		gtk_box_pack_start (GTK_BOX (chooser->view_win->top_box),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);

	if (filter_label->allocation.width > section_label->allocation.width) {
		gtk_widget_size_request (filter_label, &req);
		gtk_widget_set_usize (section_label, req.width, -1);
	}
	else {
		gtk_widget_size_request (section_label, &req);
		gtk_widget_set_usize (filter_label, req.width, -1);
	}
	
	// section combobox changed signal
	g_signal_connect (G_OBJECT(chooser->combobox), "changed",
		G_CALLBACK(combobox_changed), chooser);
	
	// load application data
	add_apps_from_directory (chooser, app_path);
	app = chooser->apps->data;
	
	// selected application name
	chooser->name_label = gtk_label_new(app->name);
	gtk_label_set_line_wrap(GTK_LABEL(
		chooser->name_label), TRUE);
	gtk_label_set_justify(GTK_LABEL(
		chooser->name_label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(
		chooser->name_label), PANGO_ELLIPSIZE_END);
	markup = g_markup_printf_escaped (
		"<span size=\"large\" weight=\"bold\">%s</span>",
		GTK_LABEL (chooser->name_label)->text);
	gtk_label_set_markup (GTK_LABEL (chooser->name_label), markup);
	g_free (markup);
	gtk_box_pack_start (GTK_BOX (chooser->view_win->bottom_box),
		chooser->name_label, FALSE, FALSE, 5);
	gtk_widget_show(chooser->name_label);
	
	// selected application description
	chooser->description_label = gtk_label_new(app->description);
	gtk_label_set_line_wrap(GTK_LABEL(chooser->description_label), TRUE);
	gtk_label_set_justify(GTK_LABEL(
		chooser->description_label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(
		chooser->description_label), PANGO_ELLIPSIZE_END);
	gtk_box_pack_start (GTK_BOX (chooser->view_win->bottom_box),
		chooser->description_label, FALSE, FALSE, 10);
	gtk_widget_show(chooser->description_label);
	
	// connect to some icon_view signals
	g_signal_connect (G_OBJECT(chooser->view_win->win), "destroy",
		G_CALLBACK(destroy_cb), chooser);
	g_signal_connect (GTK_OBJECT(chooser->view_win->icon_view), "child-selected",
		G_CALLBACK(icon_view_child_selected), chooser);
	g_signal_connect_after(G_OBJECT (chooser->view_win->win), "expose-event",
		G_CALLBACK(kiba_icon_view_expose_cb), chooser);
	
	gtk_widget_grab_focus (chooser->filter);
	
	// default show all icons
	chooser->active_section = KIBA_APP_SECTION_ALL;
	
	// add applications to the view in a timeout
	// avoids gui blocking
	chooser->currently_processing = chooser->apps;
	chooser->timeout_id = g_timeout_add(100, process_apps, chooser);

	return chooser;
}
