/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>

#include "kiba-utils.h"
#include "kiba-icon-view-win.h"

#include "marshallers.h"

#define VIEW_WIN_COLUMNS(view_win) kiba_icon_view_get_columns(KIBA_ICON_VIEW(view_win->icon_view))
#define KIBA_ICON_VIEW_MAX_ICONS(view_win)  view_win->rows*VIEW_WIN_COLUMNS(view_win)

G_DEFINE_TYPE (KibaIconViewWin, kiba_icon_view_win, G_TYPE_OBJECT)

static void
kiba_icon_view_win_destroy_cb (GtkWidget *widget,
		gpointer data)
{
	KibaIconViewWin *view_win = data;
	g_object_unref (view_win);
}

static gboolean
kiba_icon_view_win_button_release_cb (GtkWidget  *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaIconViewWin *view_win = (KibaIconViewWin*) data;

	if (event->button != 2) {
		return FALSE;
	}

	view_win->win = NULL;
	gtk_widget_destroy (widget);

	return TRUE;
}

static void
kiba_icon_view_size_allocate_cb (GtkWidget *widget,
		GtkAllocation *allocation,
		gpointer data)
{
	gtk_widget_queue_draw (((KibaIconViewWin*) data)->win);
}

static void
icon_view_close_cb (GtkWidget *widget,
		GdkEventButton *button,
		gpointer user_data)
{
	KibaIconViewWin *view_win = user_data;
	gtk_widget_destroy (view_win->win);
}

/// public methods

/**
 * kiba_icon_view_win_add:
 * @view_win: The KibaIconViewWin instance.
 * @widget: The widget to add.
 *
 * Adds widget to the grid of view_win.
 */
void
kiba_icon_view_win_add (KibaIconViewWin *view_win,
		GtkWidget *widget)
{
	KibaIconView *icon_view;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW_WIN(view_win));
	g_return_if_fail (GTK_IS_WIDGET(widget));
	
	icon_view = KIBA_ICON_VIEW(view_win->icon_view);
	
	gtk_container_add (GTK_CONTAINER(icon_view), widget);
}

/**
 * kiba_icon_view_win_remove:
 * @view_win: The KibaIconViewWin instance.
 * @widget: The widget to remove.
 *
 * Removes widget from view_win's grid.
 */
void
kiba_icon_view_win_remove (KibaIconViewWin *view_win,
		GtkWidget *widget)
{
	KibaIconView *icon_view;
	gint active, before, after;

	g_return_if_fail (KIBA_IS_ICON_VIEW_WIN(view_win));

	icon_view = KIBA_ICON_VIEW(view_win->icon_view);

	gtk_container_remove (GTK_CONTAINER(icon_view), widget);

	// resize immediately
	GtkAllocation alloc;
	alloc = view_win->icon_view->allocation;
	gtk_widget_queue_resize (view_win->icon_view);
	gtk_widget_size_allocate(view_win->icon_view, &alloc);
}

/// class implementation

static void
kiba_icon_view_win_class_init (KibaIconViewWinClass *class)
{
}

static void
kiba_icon_view_win_init (KibaIconViewWin *view_win)
{
}

/**
 * kiba_icon_view_win_new:
 * @position: The position of the boxes for custom widgets.
 * @columns: The number of columns in the grid.
 * @rows: The number of rows in the grid.
 * @close_button: TRUE if a close button should be added, otherwise FALSE
 * @use_scrolled_window: use scrolled window for icon view?
 * @attached: attached to an object on the dock?
 *
 * Creates and returns a KibaIconViewWin instance.
 *
 * Returns: The KibaIconViewWin instance created.
 */
KibaIconViewWin*
kiba_icon_view_win_new (KibaIconViewBoxPosition position,
		gint columns,
		gint rows,
		gboolean close_button,
		gboolean use_scrolled_window,
		gboolean attached)
{
	KibaIconViewWin *view_win;
	GtkWidget *vbox;
	GtkWidget *mainbox;
	GtkWidget *arrow;
	GtkWidget *image;
	GtkWidget *event_box;
	GtkWidget *scroll;
	
	view_win = g_object_new (KIBA_TYPE_ICON_VIEW_WIN, NULL);
	
	view_win->rows = rows;
	
	view_win->win = kiba_info_win_new(attached);
	gtk_window_set_decorated (GTK_WINDOW(view_win->win), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(view_win->win), FALSE);

	mainbox = gtk_hbox_new (FALSE, 0);
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show (vbox);

	view_win->top_box    = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), view_win->top_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->top_box);

	view_win->right_box  = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(view_win->top_box), view_win->right_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->right_box);

	view_win->bottom_box = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_end (GTK_BOX(vbox), view_win->bottom_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->bottom_box);
	
	if (use_scrolled_window) {
		scroll = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scroll),
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	}
	view_win->icon_view = kiba_icon_view_new();
	kiba_icon_view_set_columns (KIBA_ICON_VIEW(view_win->icon_view), columns);
	gtk_widget_show (view_win->icon_view);
	if (use_scrolled_window) {
		gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll), view_win->icon_view);
		gint spacing = gtk_box_get_spacing (GTK_BOX(view_win->icon_view));
		gint border_width =
			border_width = gtk_container_get_border_width (GTK_CONTAINER(view_win->icon_view));
		gtk_widget_set_usize(scroll,
				(48+spacing)*columns + spacing + 2*border_width + 24, 340);
		gtk_widget_show(scroll);
		gtk_box_pack_start (GTK_BOX(vbox), scroll, FALSE, FALSE, 0);
	} else {
		gtk_box_pack_start (GTK_BOX(vbox), view_win->icon_view, FALSE, FALSE, 0);
	}

	view_win->button_box = gtk_hbox_new (FALSE, 0);
	if (close_button) {
		event_box = gtk_event_box_new ();
		gtk_event_box_set_visible_window (GTK_EVENT_BOX(event_box), FALSE);
		image = gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_widget_set_usize (event_box, 24, 24);
		gtk_container_add (GTK_CONTAINER (event_box), image);
		gtk_box_pack_end (GTK_BOX(view_win->button_box), event_box, FALSE, FALSE, 5);
		gtk_widget_show (image);
		gtk_widget_show (event_box);
		g_signal_connect (G_OBJECT(event_box), "button-press-event",
			G_CALLBACK (icon_view_close_cb), view_win);
	}

	if (close_button)
		gtk_widget_show (view_win->button_box);
	gtk_box_pack_start (GTK_BOX(view_win->right_box), view_win->button_box, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER(view_win->win), mainbox);
	gtk_widget_show (mainbox);

	g_signal_connect (G_OBJECT(view_win->win), "button-release-event",
		G_CALLBACK (kiba_icon_view_win_button_release_cb), view_win);
	g_signal_connect_after (G_OBJECT(view_win->win), "destroy",
		G_CALLBACK (kiba_icon_view_win_destroy_cb), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "size-allocate",
		G_CALLBACK(kiba_icon_view_size_allocate_cb), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "size-allocate",
		G_CALLBACK(kiba_icon_view_size_allocate_cb), view_win);

	return view_win;
}
