/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-types.h"
#include "kiba-cairo-utils.h"
#include "kiba-tooltip.h"

G_DEFINE_TYPE (KibaTooltip, kiba_tooltip, GTK_TYPE_WINDOW)

#define KIBA_TOOLTIP_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
		KIBA_TYPE_TOOLTIP, KibaTooltipPrivate))

// preference keys
#define PREFERENCES_FONT_KEY "tooltip_font"
#define PREFERENCES_BORDER_WIDTH_KEY "tooltip_border_width"
#define PREFERENCES_GRADIENT_KEY "tooltip_gradient"
#define PREFERENCES_BORDER_RGBA_KEY "tooltip_border_rgba"
#define PREFERENCES_TEXT_RGBA_KEY "tooltip_text_rgba"
#define PREFERENCES_BACKGROUND_KEY "tooltip_background"
#define PREFERENCES_BORDER_STYLE_KEY "tooltip_border_style"

// preference defaults
#define PREFERENCES_FONT_DEFAULT "Sans 12"
#define PREFERENCES_BORDER_WIDTH_DEFAULT "5"
#define PREFERENCES_GRADIENT_DEFAULT "[#082D7688]"
#define PREFERENCES_BORDER_RGBA_DEFAULT "#000000FF"
#define PREFERENCES_TEXT_RGBA_DEFAULT "#000000FF"
#define PREFERENCES_BACKGROUND_DEFAULT "gradient"
#define PREFERENCES_BORDER_STYLE_DEFAULT "rounded"

static PrefMember tooltip_options[] = {
	{ PREFERENCES_FONT_KEY,
		PREFERENCES_FONT_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BORDER_WIDTH_KEY,
		PREFERENCES_BORDER_WIDTH_DEFAULT, KIBA_INT },
	{ PREFERENCES_GRADIENT_KEY,
		PREFERENCES_GRADIENT_DEFAULT, KIBA_LIST },
	{ PREFERENCES_BORDER_RGBA_KEY,
		PREFERENCES_BORDER_RGBA_DEFAULT, KIBA_STRING },
	{ PREFERENCES_TEXT_RGBA_KEY,
		PREFERENCES_TEXT_RGBA_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BACKGROUND_KEY,
		PREFERENCES_BACKGROUND_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BORDER_STYLE_KEY,
		PREFERENCES_BORDER_STYLE_DEFAULT, KIBA_STRING }
};

typedef struct {
	PangoLayout *layout;
	char *currentText;
} KibaTooltipPrivate;

/** Private Methods **/

static void __update_size (KibaTooltip *tooltip) {
	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);
	gint borderWidth = gtk_container_get_border_width(GTK_CONTAINER(tooltip));
	PangoRectangle  log;

	if (priv->layout)
		pango_layout_get_pixel_extents (priv->layout, NULL, &log);
	else {
		log.width = 0;
		log.height = 0;
	}

	gtk_widget_set_usize(GTK_WIDGET(tooltip),
		log.width + 2*borderWidth,
		log.height + 2*borderWidth);
}

static void __update_layout (KibaTooltip *tooltip) {
	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);
	KibaWin *win = kiba_get_win(kiba_get());
	cairo_t *cr = NULL;

	cr = gdk_cairo_create (GTK_WIDGET(win)->window);
	cairo_update_pango_layout (priv->currentText,
		tooltip->font,
		cr,
		&priv->layout);
	cairo_destroy (cr);
	__update_size(tooltip);
}

/** Public Methods **/

void kiba_tooltip_set_text(KibaTooltip *tooltip,
                           const char *text) {
	g_return_if_fail(KIBA_IS_TOOLTIP(tooltip));
	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);

	if (priv->currentText == NULL) {
		if (text == NULL)
			return;
	}
	else {
		if (text != NULL && strcmp(text, priv->currentText) == 0)
			return;
		g_free (priv->currentText);
		priv->currentText = NULL;
	}

	if (text != NULL)
		priv->currentText = g_strdup(text);

	__update_layout(tooltip);
}

void kiba_tooltip_set_font (KibaTooltip *tooltip,
                            const char *font) {
	g_return_if_fail(KIBA_IS_TOOLTIP(tooltip));
	g_return_if_fail(font != NULL);

	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);

	if (tooltip->font)
		g_free(tooltip->font);
	tooltip->font = g_strdup(font);

	if (priv->currentText != NULL) {
		__update_layout(tooltip);
		__update_size(tooltip);
		gtk_widget_queue_draw(GTK_WIDGET(tooltip));
	}
}

void kiba_tooltip_set_border_width (KibaTooltip *tooltip,
                                    gint border_width) {
	g_return_if_fail(KIBA_IS_TOOLTIP(tooltip));

	gtk_container_set_border_width(GTK_CONTAINER(tooltip), border_width);
	__update_size(tooltip);
}

void kiba_tooltip_set_gradient (KibaTooltip *tooltip,
                                KibaRGBA    *rgba,
                                gint         numColors) {
	g_return_if_fail(KIBA_IS_TOOLTIP(tooltip));
	g_return_if_fail(numColors > 0);
	g_return_if_fail(rgba != NULL);
	gint i;

	tooltip->numColors = numColors;

	g_free(tooltip->gradient);
	tooltip->gradient = rgba;

	gtk_widget_queue_draw(GTK_WIDGET(tooltip));
}

/*** Widget Class Implementations ***/

/*
 * gtk_widget_expose implementation.
 */
static gboolean kiba_tooltip_expose_event (GtkWidget      *widget,
				       GdkEventExpose *expose) {
	KibaTooltip *tooltip = KIBA_TOOLTIP(widget);
	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);
	gint borderWidth = gtk_container_get_border_width(GTK_CONTAINER(widget));
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern;

	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;

	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);

	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint(cr);

	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

	cairo_paint_shape (cr,
			  tooltip->cornerStyle, tooltip->cornerStyle,
			  tooltip->cornerStyle, tooltip->cornerStyle,
			  borderWidth/2,
			  borderWidth/2,
			  widget->allocation.width - borderWidth/2,
			  widget->allocation.height - borderWidth/2,
			  6.0);

	if (!tooltip->file)
		pattern = cairo_get_gradient_pattern (borderWidth/2, borderWidth/2,
				widget->allocation.width - borderWidth,
				widget->allocation.height - borderWidth,
				tooltip->gradient, tooltip->numColors);
	else
		pattern = cairo_get_png_pattern (tooltip->file, 1.0);

	cairo_set_source (cr, pattern);
	cairo_fill_preserve (cr);
	cairo_pattern_destroy (pattern);

	cairo_set_source_rgba (cr,
		tooltip->borderColor.r, tooltip->borderColor.g,
		tooltip->borderColor.b, tooltip->borderColor.a);
	cairo_set_line_width (cr, borderWidth);
	cairo_stroke (cr);

	cairo_translate (cr, borderWidth, borderWidth);
	cairo_set_source_rgba (cr,
		tooltip->textColor.r, tooltip->textColor.g,
		tooltip->textColor.b, tooltip->textColor.a);
	pango_cairo_show_layout (cr, priv->layout);

	cairo_destroy (cr);

	return TRUE;
}

/*
 * gtk_widget_screen_changed implementation.
 */
static void kiba_tooltip_screen_changed (GtkWidget *widget,
                                     GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;

	screen = gtk_widget_get_screen (widget);

	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}

	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

/*** Class init ***/

static void kiba_tooltip_class_init (KibaTooltipClass *class) {
	GtkWidgetClass    *widget_class;
	GtkContainerClass *container_class;

	widget_class    = GTK_WIDGET_CLASS (class);
	container_class = GTK_CONTAINER_CLASS (class);

	g_type_class_add_private (class, sizeof (KibaTooltipPrivate));

	widget_class->expose_event = kiba_tooltip_expose_event;
	widget_class->screen_changed = kiba_tooltip_screen_changed;
}

static void kiba_tooltip_init (KibaTooltip *tooltip) {
	KibaTooltipPrivate *priv = KIBA_TOOLTIP_GET_PRIVATE(tooltip);

	gtk_widget_add_events (GTK_WIDGET (tooltip), GDK_EXPOSURE_MASK);

	kiba_tooltip_screen_changed (GTK_WIDGET (tooltip), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (tooltip), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (tooltip), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (tooltip), FALSE);
	gtk_window_set_title (GTK_WINDOW (tooltip), PACKAGE);
	gtk_window_set_type_hint (GTK_WINDOW(tooltip), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_window_set_keep_above (GTK_WINDOW(tooltip), TRUE);
	gtk_window_stick (GTK_WINDOW(tooltip));
	gtk_container_set_border_width(GTK_CONTAINER(tooltip), 2);
	kiba_tooltip_screen_changed(GTK_WIDGET(tooltip), NULL);

	tooltip->gradient = g_new0(KibaRGBA, 1);
	tooltip->gradient[0] = (KibaRGBA) {1.0,1.0,1.0,1.0};
	tooltip->numColors = 1;
	tooltip->borderColor = (KibaRGBA) {0.0,0.0,0.0,1.0};
	tooltip->textColor = (KibaRGBA) {0.0,0.0,0.0,1.0};
	tooltip->file = NULL;
	tooltip->cornerStyle = KIBA_ROUNDED;
	tooltip->font = g_strdup("Sans 12");

	priv->layout = NULL;
	priv->currentText = NULL;
}

/**
 * kiba_tooltip_new:
 *
 * Creates and returns a KibaTooltip Instance.
 *
 * Returns: the KibaTooltip Instance created.
 */
KibaTooltip* kiba_tooltip_new (void) {
	return g_object_new (KIBA_TYPE_TOOLTIP, NULL);
}

// TODO: add settings and callbacks
/*
static void tooltip_settings_cb (const char *option_id,
				  gpointer    value,
				  gpointer    data) {
	Kiba *kiba = kiba_get();
	KibaTooltip *tooltip;

	tooltip = kiba_get_tooltip(kiba);

	if (strcmp (option_id, "tooltip_gradient") == 0) {
		GSList *list = value;
		gint length = g_slist_length(list);
		KibaRGBA *gradient = g_new0(KibaRGBA, length);
		kiba_prefs_string_list_to_rgba_array(list, gradient);
		kiba_tooltip_set_gradient(tooltip, gradient, length);
	}
	else if (strcmp (option_id, "tooltip_font") == 0) {
		kiba_tooltip_set_font(tooltip, (char*) value);
	}
	else if (strcmp (option_id, "tooltip_border_width") == 0) {
		kiba_tooltip_set_border_width(tooltip, GPOINTER_TO_INT(value));
	}
	else if (strcmp (option_id, "tooltip_border_rgba") == 0) {
		tooltip->borderColor = kiba_prefs_rgba_from_string ((char*) value);
		gtk_widget_queue_draw(GTK_WIDGET(tooltip));
	}
	else if (strcmp (option_id, "tooltip_text_rgba") == 0) {
		tooltip->textColor = kiba_prefs_rgba_from_string ((char*) value);
		gtk_widget_queue_draw(GTK_WIDGET(tooltip));
	}
	else if (strcmp (option_id, "tooltip_background") == 0) {
		if (g_str_has_suffix ((char*) value, ".png") || g_str_has_suffix ((char*) value, ".svg"))
			tooltip->file = (char*) value;
		else
			tooltip->file = NULL;
		gtk_widget_queue_draw(GTK_WIDGET(tooltip));
	}
	else if (strcmp (option_id, "tooltip_border_style") == 0) {
		tooltip->cornerStyle = kiba_prefs_edge_style_from_string ((char*) value);
		gtk_widget_queue_draw(GTK_WIDGET(tooltip));
	}
}

	kiba_prefs_add_callback (generalFile, "tooltip_gradient",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_font",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_border_width",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_border_rgba",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_text_rgba",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_background",
		tooltip_settings_cb, kiba, TRUE);
	kiba_prefs_add_callback (generalFile, "tooltip_border_style",
		tooltip_settings_cb, kiba, TRUE);
*/
